
-------------------------------------------------------------------------------
--
-- Simple VGA raster display
--
-- Stephen A. Edwards
-- sedwards@cs.columbia.edu
--
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity wii_vga is
  
  port (
    clk        : in  std_logic;
    reset_n    : in  std_logic;
    read       : in  std_logic;
    write      : in  std_logic;
    chipselect : in  std_logic;
    address    : in  unsigned(15 downto 0);
    readdata   : out unsigned(15 downto 0);
    writedata  : in  unsigned(15 downto 0);

    VGA_CLK,                         		 -- Clock
    VGA_HS,                          		 -- H_SYNC
    VGA_VS,                          		 -- V_SYNC
    VGA_BLANK,                       		 -- BLANK
    VGA_SYNC : out std_logic;        		 -- SYNC
    VGA_R,                           		 -- Red[9:0]
    VGA_G,                           		 -- Green[9:0]
    VGA_B : out std_logic_vector(9 downto 0) -- Blue[9:0]
    );

end wii_vga;

architecture rtl of wii_vga is
  
  -- Video parameters
  constant HTOTAL       : integer := 800;
  constant HSYNC        : integer := 96;
  constant HBACK_PORCH  : integer := 48;
  constant HACTIVE      : integer := 640;
  constant HFRONT_PORCH : integer := 16;
  
  constant VTOTAL       : integer := 525;
  constant VSYNC        : integer := 2;
  constant VBACK_PORCH  : integer := 33;
  constant VACTIVE      : integer := 480;
  constant VFRONT_PORCH : integer := 10;

  -- Signals for the video controller
  signal Hcount, sprite_x : unsigned(15 downto 0);  -- Horizontal position (0-800)
  signal Vcount, sprite_y : unsigned(15 downto 0);  -- Vertical position (0-524)
  signal EndOfLine, EndOfField : std_logic;
  signal tmp_vga_r : std_logic_vector(9 downto 0) := "0000000000";
  signal tmp_vga_g : std_logic_vector(9 downto 0) := "0000000000";
  signal tmp_vga_b : std_logic_vector(9 downto 0) := "0000000000";

  -- Sync. signals
  signal vga_hblank, vga_hsync, vga_vblank, vga_vsync : std_logic;  

  -- Game VGA signals
  signal draw_char, draw_obj, draw_maze, draw_goal : std_logic;

  signal which_maze_tile : integer := 0;						-- which of the 4 actual maze tiles to use
  signal maze_x, maze_y  : integer := 0;						-- rotation-adjusted indexes into that maze tile
  signal rect_maze_h, rect_maze_v : std_logic;					-- are we currently within the bounds of the maze tile?
  signal maze_tile_x, maze_tile_y : integer := 0;				-- indexes into the maze_desc array  
  signal valid_maze_tile : std_logic;							-- is this the empty tile?

  signal which_obj_tile : integer := 0;						-- which of the 4 actual obj tiles to use
  signal obj_x, obj_y  : integer := 0;						-- rotation-adjusted indexes into that obj tile
  signal rect_obj_h, rect_obj_v : std_logic;				-- are we currently within the bounds of the obj tile?
  signal obj_tile_x, obj_tile_y : integer := 0;				-- indexes into the obj_desc array
  signal can_show_cur_color : std_logic;					-- allowed to show the obj of the color currently being drawn?
  signal valid_obj_tile : std_logic;						-- is this the empty tile?
  
  signal rect_goal_h, rect_goal_v : std_logic;				-- are we currently within the bounds of the goal tile?

  signal which_char_tile : unsigned(15 downto 0) := "0000000000000000";	-- which character tile to use
  signal tmp_char_tile : unsigned(15 downto 0) :=   "0000000000000000";	-- which character tile to use (temp)
  signal char_x, char_y  : integer := 0;								-- indexes into that char tile
  signal rect_char_h, rect_char_v : std_logic;							-- are we currently within the bounds of the char tile?

  signal count_x, count_y : unsigned(4 downto 0) := "00000";						-- unrotated indexes into maze and object arrays
  signal tmp_count_x, tmp_count_y : unsigned(15 downto 0) := "0000000000000000";	-- unrotated indexes into maze and object arrays (temp)
  signal anim_count   : integer := 0;												-- a count of clock cycles so that we can animate at a reasonable speed

  -- Game parameters
  constant TILE_SIZE        : integer := 32;	     							-- the size of a tile in pixels
  constant MAX_MAZE_SIZE_X  : integer := 16;       	 							-- the size of the maze in number of tiles
  constant MAX_MAZE_SIZE_Y  : integer := 16;									-- the size of the maze in number of tiles
  constant EXTRA_LEFT		: unsigned(15 downto 0) := to_unsigned(HTOTAL,16);	-- The extra amount of screen available at the left edge of the physical screen
  constant EXTRA_TOP		: unsigned(15 downto 0) := to_unsigned(VTOTAL,16);  -- The extra amount of screen available at the top edge of the physical screen

  -- Actual game tiles
  type tile_type is array(0 to TILE_SIZE-1) of unsigned(0 to TILE_SIZE-1);
  type maze_tile_type is array(0 to 3) of tile_type;
  constant MAZE_TILES : maze_tile_type := 
								( ( "11111000000000000000000000000000",
									"11111000000000000000000000000000",
									"11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
									"11111000000000000000000000000000",
									"11111000000000000000000000000000" ), 
									
							      ( "11111111111111111111111111111111",
									"11111111111111111111111111111111",
									"11111111111111111111111111111111",
					                "11111111111111111111111111111111",
					                "11111111111111111111111111111111",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
									"11111000000000000000000000000000",
									"11111000000000000000000000000000" ), 
									
							      ( "11111000000000000000000000011111",
									"11111000000000000000000000011111",
									"11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
									"11111000000000000000000000011111",
									"11111000000000000000000000011111" ), 
									
								  ( "11111000000000000000000000011111",
									"11111000000000000000000000011111",
									"11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111000000000000000000000011111",
					                "11111111111111111111111111111111",
					                "11111111111111111111111111111111",
					                "11111111111111111111111111111111",
									"11111111111111111111111111111111",
									"11111111111111111111111111111111" ) );
									
  type obj_tile_type is array(0 to 1) of tile_type;
  constant OBJ_TILES : obj_tile_type := 
								( ( "11111000000000000000000000000000",
									"11111000000000000000000000000000",
									"11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
					                "11111000000000000000000000000000",
									"11111000000000000000000000000000",
									"11111000000000000000000000000000" ), 
									
							      ( "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
								    "00000000000000111100000000000000",
					                "00000000000001111110000000000000",
					                "00000000000001100110000000000000",
					                "00000000000001100110000000000000",
					                "00000000000001111110000000000000",
					                "00000000000000111100000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011110000000000000",
					                "00000000000000011110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011111000000000000",
					                "00000000000000011111000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000" ) ); 								

  type char_tile_type is array(0 to 3) of tile_type;
  constant CHAR_TILES : char_tile_type := 
								( ( "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000011111111000000000000",
					                "00000000000011111111000000000000",
					                "00000000000001111110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000011100011000111000000000",
					                "00000000000111011011100000000000",
					                "00000000000001111110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000111000011100000000000",
					                "00000000011110000001111000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000" ), 
									
							      ( "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000011111111000000000000",
					                "00000000000011111111000000000000",
					                "00000000000001111110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000111111111111111100000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000111000011100000000000",
					                "00000000011110000001111000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000" ),  
									
							     ( "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000011111111000000000000",
					                "00000000000011111111000000000000",
					                "00000000000001111110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000011100011000111000000000",
					                "00000000000111011011100000000000",
					                "00000000000001111110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000111000011100000000000",
					                "00000000011110000001111000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000" ), 
									
							      ( "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000011111111000000000000",
					                "00000000000011111111000000000000",
					                "00000000000001111110000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000111111111111111100000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000000011000000000000000",
					                "00000000000001111110000000000000",
					                "00000000000111000011100000000000",
					                "00000000011110000001111000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000" ) );
									
  constant GOAL_TILE : tile_type := 
								  ( "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000111111111111111100000000",
					                "00000001111111111111111110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001111111111111111110000000",
					                "00000001111111111111111110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001100000011000000110000000",
					                "00000001111111111111111110000000",
					                "00000000111111111111111100000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
					                "00000000000000000000000000000000",
									"00000000000000000000000000000000",
									"00000000000000000000000000000000" );

  -- Virtual Tiles

  -- Virtual Maze tiles
  -- Bit 7-6: index into MAZE_TILES array
  -- Bit 5-4: rotation (00 = none, 01 = 90 c, 10 = 180 c, 11 = 270 c)
  -- Bit 3:   allows player to move left? (1 if affirmative)          NO LONGER USED
  -- Bit 2:   allows player to move right? (1 if affirmative)         NO LONGER USED
  -- Bit 1:   allows player to move up? (1 if affirmative)            NO LONGER USED
  -- Bit 0:   allows player to move down? (1 if affirmative)          NO LONGER USED
  type virtual_maze_tile_type is array(0 to 14) of unsigned(7 downto 0);
  constant VIRTUAL_MAZE_TILES : virtual_maze_tile_type := ( "00000111",
															"00011101",
															"00101011",
															"00111110",
															"01000101",
															"01011001",
															"01101010",
															"01110110",
															"10000011",
															"10011100",
															"11000010",
															"11010100",
															"11100001",
															"11111000",
															"11111111" );
															
	constant LEFT_WALL 			 : unsigned(3 downto 0) := "0000";
	constant TOP_WALL 			 : unsigned(3 downto 0) := "0001";
	constant RIGHT_WALL 		 : unsigned(3 downto 0) := "0010";
	constant BOTTOM_WALL		 : unsigned(3 downto 0) := "0011";
	constant TOP_LEFT_CORNER 	 : unsigned(3 downto 0) := "0100";
	constant TOP_RIGHT_CORNER	 : unsigned(3 downto 0) := "0101";
	constant BOTTOM_RIGHT_CORNER : unsigned(3 downto 0) := "0110";
	constant BOTTOM_LEFT_CORNER  : unsigned(3 downto 0) := "0111";
	constant VERTICAL_HALL 		 : unsigned(3 downto 0) := "1000";
	constant HORIZONTAL_HALL 	 : unsigned(3 downto 0) := "1001";
	constant BOTTOM_DEADEND 	 : unsigned(3 downto 0) := "1010";
	constant LEFT_DEADEND 		 : unsigned(3 downto 0) := "1011";
	constant TOP_DEADEND 		 : unsigned(3 downto 0) := "1100";
	constant RIGHT_DEADEND 		 : unsigned(3 downto 0) := "1101";
	constant NO_MAZE_TILE		 : unsigned(3 downto 0) := "1110";
						
  -- Virtual object tiles
  -- Bit 8:   index into OBJ_TILES array
  -- Bit 7-6: rotation (00 = none, 01 = 90 c, 10 = 180 c, 11 = 270 c)
  -- Bit 5-4: color (00 = red, 01 = green, 10 = blue, 11 = invalid)
  -- Bit 3:   allows player to move left? (1 if affirmative)           NO LONGER USED
  -- Bit 2:   allows player to move right? (1 if affirmative)          NO LONGER USED
  -- Bit 1:   allows player to move up? (1 if affirmative)             NO LONGER USED
  -- Bit 0:   allows player to move down? (1 if affirmative)           NO LONGER USED
  type virtual_obj_tile_type is array(0 to 15) of unsigned(8 downto 0);
  constant VIRTUAL_OBJ_TILES : virtual_obj_tile_type := ( "000000111",
														  "001001101",
														  "010001011",
														  "011001110",
														  "000010111",
														  "001011101",
														  "010011011",
														  "011011110",
														  "000100111",
														  "001101101",
														  "010101011",
														  "011101110",
														  "100001111",
														  "100011111",
														  "100101111",
														  "111111111" );
														
	constant LEFT_WALL_RED 			 : unsigned(3 downto 0) := "0000";
	constant TOP_WALL_RED 			 : unsigned(3 downto 0) := "0001";
	constant RIGHT_WALL_RED 		 : unsigned(3 downto 0) := "0010";
	constant BOTTOM_WALL_RED		 : unsigned(3 downto 0) := "0011";
	constant LEFT_WALL_GREEN 	 	 : unsigned(3 downto 0) := "0100";
	constant TOP_WALL_GREEN	 	     : unsigned(3 downto 0) := "0101";
	constant RIGHT_WALL_GREEN 		 : unsigned(3 downto 0) := "0110";
	constant BOTTOM_WALL_GREEN  	 : unsigned(3 downto 0) := "0111";
	constant LEFT_WALL_BLUE 		 : unsigned(3 downto 0) := "1000";
	constant TOP_WALL_BLUE 	 		 : unsigned(3 downto 0) := "1001";
	constant RIGHT_WALL_BLUE	 	 : unsigned(3 downto 0) := "1010";
	constant BOTTOM_WALL_BLUE 		 : unsigned(3 downto 0) := "1011";
	constant KEY_RED 		 		 : unsigned(3 downto 0) := "1100";
	constant KEY_GREEN 		 		 : unsigned(3 downto 0) := "1101";
	constant KEY_BLUE		 		 : unsigned(3 downto 0) := "1110";	
	constant NO_OBJ_TILE			 : unsigned(3 downto 0) := "1111";
														
  -- Maze and object descriptors
  type desc_sub_type is array(0 to MAX_MAZE_SIZE_X-1) of unsigned(3 downto 0);
  type desc_type is array(0 to MAX_MAZE_SIZE_Y-1) of desc_sub_type;
  signal maze_desc : desc_type := ( (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
									(LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND),
								    (LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND, LEFT_DEADEND) );

  signal obj_desc : desc_type :=  ( (NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE),
									(NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE, NO_OBJ_TILE) );
									
  -- Other game signals
  signal which_keys      : unsigned(15 downto 0) := "0000000000000111";   -- which keys haven't been collected (first bit is red, second is blue, third is green)

  signal maze_size_x      : unsigned(3 downto 0) := to_unsigned(MAX_MAZE_SIZE_X,4);          		-- the size of the maze in number of tiles
  signal maze_size_y      : unsigned(3 downto 0) := to_unsigned(MAX_MAZE_SIZE_Y,4);          		-- the size of the maze in number of tiles
  signal maze_size_pix_x  : unsigned(15 downto 0) := to_unsigned(MAX_MAZE_SIZE_X * TILE_SIZE,16);	-- the size of the maze in pixels (not settable by software, but should be set by the hardware if the maze_size is set)
  signal maze_size_pix_y  : unsigned(15 downto 0) := to_unsigned(MAX_MAZE_SIZE_Y * TILE_SIZE,16);	-- the size of the maze in pixels (not settable by software, but should be set by the hardware if the maze_size is set)
  signal maze_x_offset, tmp_maze_x_offset  : unsigned(15 downto 0) := to_unsigned(HSYNC + HBACK_PORCH,16)+EXTRA_LEFT;    -- left position of beginning of maze in pixels ( including SYNC and BACK_PORCH)
  signal maze_y_offset, tmp_maze_y_offset  : unsigned(15 downto 0) := to_unsigned(VSYNC + VBACK_PORCH,16)+EXTRA_TOP;  	 -- top position of the beginning of the maze

  constant window_x_center	 : unsigned(15 downto 0) := to_unsigned(HSYNC+HBACK_PORCH+320,16);
  constant window_y_center	 : unsigned(15 downto 0) := to_unsigned(VSYNC+VBACK_PORCH+240,16);
  signal half_window_size_x  : unsigned(7 downto 0) := "00000010";          -- horiz size of window in which you can see the maze in # of tiles
  signal half_window_size_y  : unsigned(7 downto 0) := "00000010";          -- verti size of window in which you can see the maze in # of tiles
  signal is_on_window : std_logic := '0';
  signal is_in_window : std_logic := '0';

  signal end_xpos_maze   : unsigned(15 downto 0) := to_unsigned(MAX_MAZE_SIZE_X-1,16);
  signal end_ypos_maze   : unsigned(15 downto 0) := to_unsigned(MAX_MAZE_SIZE_Y-1,16);

  signal cur_xpos_pix, tmp_cur_xpos_pix  : unsigned(15 downto 0) := to_unsigned(HSYNC + HBACK_PORCH,16)+EXTRA_LEFT;		-- character's current x position in pixels
  signal cur_ypos_pix, tmp_cur_ypos_pix  : unsigned(15 downto 0) := to_unsigned(VSYNC + VBACK_PORCH,16)+EXTRA_TOP;						-- character's current y position in pixels

  signal foo : unsigned(15 downto 0);
begin  


-- Horizontal and vertical counters

  HCounter : process (clk)
  begin
    if rising_edge(clk) then      
      if reset_n = '0' then
        Hcount <= (others => '0');
      elsif EndOfLine = '1' then
        Hcount <= (others => '0');
      else
        Hcount <= Hcount + 1;
      end if;      
    end if;
  end process HCounter;

  EndOfLine <= '1' when Hcount = HTOTAL - 1 else '0';
  
  VCounter: process (clk)
  begin
    if rising_edge(clk) then      
      if reset_n = '0' then
        Vcount <= (others => '0');
      elsif EndOfLine = '1' then
        if EndOfField = '1' then
          Vcount <= (others => '0');
        else
          Vcount <= Vcount + 1;
        end if;
      end if;
    end if;
  end process VCounter;

  EndOfField <= '1' when Vcount = VTOTAL - 1 else '0';

  -- State machines to generate HSYNC, VSYNC, HBLANK, and VBLANK

  HSyncGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset_n = '0' or EndOfLine = '1' then
        vga_hsync <= '1';
      elsif Hcount = HSYNC - 1 then
        vga_hsync <= '0';
      end if;
    end if;
  end process HSyncGen;
  
  HBlankGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then
        vga_hblank <= '1';
      elsif Hcount = HSYNC + HBACK_PORCH then
        vga_hblank <= '0';
      elsif Hcount = HSYNC + HBACK_PORCH + HACTIVE then
        vga_hblank <= '1';
      end if;      
    end if;
  end process HBlankGen;

  VSyncGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then
        vga_vsync <= '1';
      elsif EndOfLine ='1' then
        if EndOfField = '1' then
          vga_vsync <= '1';
        elsif Vcount = VSYNC - 1 then
          vga_vsync <= '0';
        end if;
      end if;      
    end if;
  end process VSyncGen;

  VBlankGen : process (clk)
  begin
    if rising_edge(clk) then    
      if reset_n = '0' then
        vga_vblank <= '1';
      elsif EndOfLine = '1' then
        if Vcount = VSYNC + VBACK_PORCH - 1 then
          vga_vblank <= '0';
        elsif Vcount = VSYNC + VBACK_PORCH + VACTIVE - 1 then
          vga_vblank <= '1';
        end if;
      end if;
    end if;
  end process VBlankGen;

---------------------------------------------------------------
-- counters to keep track of current non-rotated indices into
-- maze and object tiles
---------------------------------------------------------------
PreIndexCounter : process (clk)
  begin
	if rising_edge(clk) then     
      tmp_count_x <= EXTRA_LEFT - maze_x_offset;
	  tmp_count_y <= EXTRA_TOP - maze_y_offset;
    end if;
 end process PreIndexCounter;

  IndexCounter : process (clk)
  begin
    if rising_edge(clk) then     
      if reset_n = '0' or EndOfField = '1' then
		if maze_x_offset < EXTRA_LEFT then
			count_x <= tmp_count_x(4 downto 0) - 1;
		else 
			count_x <= (others => '0');
		end if;
		
		if maze_y_offset < EXTRA_TOP then
			count_y <= tmp_count_y(4 downto 0);
		else
			count_y <= (others => '0');
		end if;
		
	  else 
		if Hcount + EXTRA_LEFT > maze_x_offset and Vcount + EXTRA_TOP >= maze_y_offset then
			count_x <= count_x + 1;
		end if;
		if Vcount + EXTRA_TOP >= maze_y_offset and EndOfLine='1' then
			count_y <= count_y + 1;
			if maze_x_offset < EXTRA_LEFT then
				count_x <= tmp_count_x(4 downto 0) - 1;
			else 
				count_x <= (others => '0');
			end if;
		end if;
	  end if;
   end if;
  end process IndexCounter;

---------------------------------------------------------------
-- VGA Controller signals for maze tiles
---------------------------------------------------------------
  MazeHGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset_n = '0' or Hcount + EXTRA_LEFT = maze_x_offset or (Hcount = HSYNC+HBACK_PORCH and maze_x_offset < EXTRA_LEFT) then
        rect_maze_h <= '1';
		maze_tile_x <= 0;
      elsif Hcount + EXTRA_LEFT= maze_x_offset + maze_size_pix_x - 1 or (Hcount = HSYNC+HBACK_PORCH+HACTIVE) then
        rect_maze_h <= '0'; 
      end if;
	maze_tile_x <= to_integer((Hcount-maze_x_offset+EXTRA_LEFT) srl 5);  
    end if;
  end process MazeHGen;

  MazeVGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then       
        rect_maze_v <= '0';
	    maze_tile_y <= 0;
      elsif EndOfLine = '1' then
        if Vcount + EXTRA_TOP = maze_y_offset or (Vcount = VSYNC+VBACK_PORCH and maze_y_offset < EXTRA_TOP) then
          rect_maze_v <= '1';
        elsif Vcount + EXTRA_TOP = maze_y_offset + maze_size_pix_y - 1 or (Vcount = VSYNC+VBACK_PORCH+VACTIVE) then
          rect_maze_v <= '0';	         
		end if;
      end if;
    maze_tile_y <= to_integer((Vcount-maze_y_offset+EXTRA_TOP) srl 5);
    end if;
  end process MazeVGen;

  MazeTileInfo : process (clk)
  begin
	if rising_edge(clk) then
      if reset_n = '0' then
		which_maze_tile <= 0;
		maze_x <= 0;
		maze_y <= 0;
	  else
		-- take first two bits of virtual tile description to get real tile #
		which_maze_tile <= to_integer(VIRTUAL_MAZE_TILES(to_integer(maze_desc(maze_tile_y)(maze_tile_x)))(7 downto 6));
		
		-- take next two to get rotation and set indices to appropriate values
		if VIRTUAL_MAZE_TILES(to_integer(maze_desc(maze_tile_y)(maze_tile_x)))(5 downto 4) = "00" then
			maze_x <= to_integer(count_x);
			maze_y <= to_integer(count_y);
		elsif VIRTUAL_MAZE_TILES(to_integer(maze_desc(maze_tile_y)(maze_tile_x)))(5 downto 4) = "01" then
			maze_x <= to_integer(count_y);
			maze_y <= to_integer(TILE_SIZE - 1 - count_x);
		elsif VIRTUAL_MAZE_TILES(to_integer(maze_desc(maze_tile_y)(maze_tile_x)))(5 downto 4) = "10" then
			maze_x <= to_integer(TILE_SIZE - 1 - count_x);
			maze_y <= to_integer(TILE_SIZE - 1 - count_y);
		elsif VIRTUAL_MAZE_TILES(to_integer(maze_desc(maze_tile_y)(maze_tile_x)))(5 downto 4) = "11" then
			maze_x <= to_integer(TILE_SIZE - 1 - count_y);
			maze_y <= to_integer(count_x);
		end if;
		
		-- check if this is the empty tile
		if maze_desc(maze_tile_y)(maze_tile_x) = NO_MAZE_TILE then
			valid_maze_tile <= '0';
		else
			valid_maze_tile <= '1';
		end if;
		
	end if;
  end if;
end process MazeTileInfo;

---------------------------------------------------------------
-- check if this is the goal tile
---------------------------------------------------------------
  GoalVHGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then       
        rect_goal_v <= '0';
	    rect_goal_h <= '0';
      else
		if maze_tile_x = end_xpos_maze and maze_tile_y = end_ypos_maze then
			rect_goal_v <= '1';
			rect_goal_h <= '1';
		else
			rect_goal_v <= '0';
			rect_goal_h <= '0';
		end if;
    end if;
 end if;
end process GoalVHGen;

---------------------------------------------------------------
-- VGA controller signals for object tiles
---------------------------------------------------------------
  ObjHGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset_n = '0' or Hcount + EXTRA_LEFT = maze_x_offset or (Hcount = HSYNC+HBACK_PORCH and maze_x_offset < EXTRA_LEFT) then
        rect_obj_h <= '1';
		obj_tile_x <= 0;
      elsif Hcount + EXTRA_LEFT = maze_x_offset + maze_size_pix_x - 1 or (Hcount = HSYNC+HBACK_PORCH+HACTIVE) then
        rect_obj_h <= '0'; 
      end if;
	obj_tile_x <= to_integer((Hcount-maze_x_offset+EXTRA_LEFT) srl 5);  
    end if;
  end process ObjHGen;

  ObjVGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then       
        rect_obj_v <= '0';
	    obj_tile_y <= 0;
      elsif EndOfLine = '1' then
        if Vcount + EXTRA_TOP = maze_y_offset or (Vcount = VSYNC+VBACK_PORCH and maze_y_offset < EXTRA_TOP)  then
          rect_obj_v <= '1';
        elsif Vcount + EXTRA_TOP = maze_y_offset + maze_size_pix_x - 1 or (Vcount = VSYNC+VBACK_PORCH+VACTIVE) then
          rect_obj_v <= '0';	         
		end if;
      end if;
    obj_tile_y <= to_integer((Vcount-maze_y_offset+EXTRA_TOP) srl 5);
    end if;
  end process ObjVGen;

  ObjTileInfo : process (clk)
  begin
	if rising_edge(clk) then
      if reset_n = '0' then
		which_obj_tile <= 0;
		obj_x <= 0;
		obj_y <= 0;
	  else
		-- take first bit of virtual tile description to get real tile #
		which_obj_tile <= to_integer(VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(8 downto 8));
		
		-- take next two to get rotation and set indices to appropriate values
		if VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(7 downto 6) = "00" then
			obj_x <= to_integer(count_x);
			obj_y <= to_integer(count_y);
		elsif VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(7 downto 6) = "01" then
			obj_x <= to_integer(count_y);
			obj_y <= to_integer(TILE_SIZE - 1 - count_x);
		elsif VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(7 downto 6) = "10" then
			obj_x <= to_integer(TILE_SIZE - 1 - count_x);
			obj_y <= to_integer(TILE_SIZE - 1 - count_y);
		elsif VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(7 downto 6) = "11" then
			obj_x <= to_integer(TILE_SIZE - 1 - count_y);
			obj_y <= to_integer(count_x);
		end if;
		
		-- take the next 2 bits to get color info
		if VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(5 downto 4) = "00" then
			tmp_vga_r <= "1111111111";
			tmp_vga_g <= "0000000000";
			tmp_vga_b <= "0000000000";
		elsif VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(5 downto 4) = "01" then
			tmp_vga_r <= "0000000000";
			tmp_vga_g <= "1111111111";
			tmp_vga_b <= "0000000000";
		elsif VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(5 downto 4) = "10" then
			tmp_vga_r <= "0000000000";
			tmp_vga_g <= "0000000000";
			tmp_vga_b <= "1111111111";
		else
			tmp_vga_r <= "0000000000";
			tmp_vga_g <= "0000000000";
			tmp_vga_b <= "0000000000";
		end if;
		
		can_show_cur_color <= which_keys(to_integer(VIRTUAL_OBJ_TILES(to_integer(obj_desc(obj_tile_y)(obj_tile_x)))(5 downto 4)));
		
		-- check if this is the empty tile
		if obj_desc(obj_tile_y)(obj_tile_x) = NO_OBJ_TILE then
			valid_obj_tile <= '0';
		else
			valid_obj_tile <= '1';
		end if;
		
	end if;
  end if;
end process ObjTileInfo;

---------------------------------------------------------------
-- VGA Controller signals for character tiles
---------------------------------------------------------------
  CharHGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset_n = '0' or Hcount + EXTRA_LEFT = cur_xpos_pix then
        rect_char_h <= '1';
      elsif Hcount + EXTRA_LEFT  = cur_xpos_pix + TILE_SIZE or (Hcount = HSYNC+HBACK_PORCH+HACTIVE) then
        rect_char_h <= '0'; 
      end if;
	char_x <= to_integer(cur_xpos_pix - EXTRA_LEFT - HCount);     --- watch out! this may go below 0 and we are dealingwith unsigned numbers
    end if;
  end process CharHGen;

  CharVGen : process (clk)
  begin
    if rising_edge(clk) then
      if reset_n = '0' then       
        rect_char_v <= '0';
      elsif EndOfLine = '1' then
        if Vcount + EXTRA_TOP = cur_ypos_pix  then
          rect_char_v <= '1';
        elsif Vcount + EXTRA_TOP = cur_ypos_pix + TILE_SIZE or (Vcount = VSYNC+VBACK_PORCH+VACTIVE) then
          rect_char_v <= '0';	         
		end if;
      end if;
    char_y <= to_integer(Vcount - cur_ypos_pix + EXTRA_TOP); 
    end if;
  end process CharVGen;

  -----------------------------------------
  -- deal with window
  -----------------------------------------
  WindowGen : process (clk)
  begin
    if rising_edge(clk) then     
      if reset_n = '0' then
        is_on_window <= '0';
		is_in_window <= '0';
      else
		if ( (Hcount = window_x_center - TILE_SIZE * half_window_size_x or Hcount = window_x_center + TILE_SIZE * half_window_size_x) and
		   (Vcount >= window_y_center - TILE_SIZE * half_window_size_y and Vcount <= window_y_center + TILE_SIZE * half_window_size_y) ) or
		   ( (Vcount = window_y_center - TILE_SIZE * half_window_size_y or Vcount = window_y_center + TILE_SIZE * half_window_size_y) and
		   (Hcount >= window_x_center - TILE_SIZE * half_window_size_x and Hcount <= window_x_center + TILE_SIZE * half_window_size_x) ) then
				is_on_window <= '1';
		else
				is_on_window <= '0';
		end if;
		
		if Hcount >= window_x_center - TILE_SIZE * half_window_size_x and Hcount <= window_x_center + TILE_SIZE * half_window_size_x and
		   Vcount >= window_y_center - TILE_SIZE * half_window_size_y and Vcount <= window_y_center + TILE_SIZE * half_window_size_y then
				is_in_window <= '1';
		else
				is_in_window <= '0';
		end if;
		
     end if;

    end if;
  end process WindowGen;

  draw_char <= rect_char_h and rect_char_v and CHAR_TILES(to_integer(which_char_tile))(char_y)(char_x) and is_in_window;
  draw_obj  <= rect_obj_h and rect_obj_v and OBJ_TILES(which_obj_tile)(obj_y)(obj_x) and can_show_cur_color and valid_obj_tile and is_in_window;
  draw_goal <= rect_goal_h and rect_goal_v and GOAL_TILE(to_integer(count_y))(to_integer(count_x)) and is_in_window;
  draw_maze <= rect_maze_h and rect_maze_v and MAZE_TILES(which_maze_tile)(maze_y)(maze_x) and valid_maze_tile and is_in_window;

  -- Registered video signals going to the video DAC
  VideoOut: process (clk, reset_n)
  begin
    if reset_n = '0' then
      VGA_R <= "0000000000";
      VGA_G <= "0000000000";
      VGA_B <= "0000000000";
    elsif rising_edge(clk) then
	  if is_on_window = '1' then
	    VGA_R <= "1111111111";
        VGA_G <= "1111111111";
        VGA_B <= "1111111111";  
	  elsif draw_char = '1' then
	    VGA_R <= "1111111111";
        VGA_G <= "1111111111";
        VGA_B <= "0000000000";
	  elsif draw_obj = '1' then
        VGA_R <= tmp_vga_r;
        VGA_G <= tmp_vga_g;
        VGA_B <= tmp_vga_b;
	  elsif draw_goal = '1' then
        VGA_R <= "0000000000";
        VGA_G <= "1111111111";
        VGA_B <= "1111111111";
	  elsif draw_maze = '1' then
		VGA_R <= "1111111111";
        VGA_G <= "1111111111";
        VGA_B <= "1111111111";
      elsif vga_hblank = '0' and vga_vblank ='0' then
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";
      else
        VGA_R <= "0000000000";
        VGA_G <= "0000000000";
        VGA_B <= "0000000000";    
      end if;
    end if;
  end process VideoOut;

  VGA_CLK <= clk;
  VGA_HS <= not vga_hsync;
  VGA_VS <= not vga_vsync;
  VGA_SYNC <= '0';
  VGA_BLANK <= not (vga_hsync or vga_vsync);


---------------------------------------------------------------
-- all communications with software from here
--------------------------------------------------------------- 
GetData: process (clk)
  begin
	if rising_edge(clk) then 
			if chipselect = '1' then
			
			    if write = '1' then
					if address = "0000000000000000" then
						tmp_maze_x_offset <= writedata;
					elsif address = "0000000000000010" then
						tmp_maze_y_offset <= writedata;	
					elsif address = "0000000000000100" then
						tmp_cur_xpos_pix <= writedata;
					elsif address = "0000000000000110" then
						tmp_cur_ypos_pix <= writedata;
					elsif address = "0000000000001000" then
						end_xpos_maze <= writedata;
					elsif address = "0000000000001010" then
						end_ypos_maze <= writedata;
					elsif address = "0000000000001100" then
						half_window_size_x <= writedata(15 downto 8);
						half_window_size_y <= writedata(7 downto 0);
					elsif address = "0000000000001110" then
						which_keys <= writedata;	
						
					elsif address = "0000000000010000" then
						maze_desc(0)(0) <= writedata(15 downto 12);
						maze_desc(0)(1) <= writedata(11 downto 8);
						maze_desc(0)(2) <= writedata(7 downto 4);
						maze_desc(0)(3) <= writedata(3 downto 0);
					elsif address = "0000000000010010" then
						maze_desc(0)(4) <= writedata(15 downto 12);
						maze_desc(0)(5) <= writedata(11 downto 8);
						maze_desc(0)(6) <= writedata(7 downto 4);
						maze_desc(0)(7) <= writedata(3 downto 0);
					elsif address = "0000000000010100" then
						maze_desc(0)(8) <= writedata(15 downto 12);
						maze_desc(0)(9) <= writedata(11 downto 8);
						maze_desc(0)(10) <= writedata(7 downto 4);
						maze_desc(0)(11) <= writedata(3 downto 0);
					elsif address = "0000000000010110" then
						maze_desc(0)(12) <= writedata(15 downto 12);
						maze_desc(0)(13) <= writedata(11 downto 8);
						maze_desc(0)(14) <= writedata(7 downto 4);
						maze_desc(0)(15) <= writedata(3 downto 0);
					
					-- second row of maze
					elsif address = "0000000000011000" then
						maze_desc(1)(0) <= writedata(15 downto 12);
						maze_desc(1)(1) <= writedata(11 downto 8);
						maze_desc(1)(2) <= writedata(7 downto 4);
						maze_desc(1)(3) <= writedata(3 downto 0);
					elsif address = "0000000000011010" then
						maze_desc(1)(4) <= writedata(15 downto 12);
						maze_desc(1)(5) <= writedata(11 downto 8);
						maze_desc(1)(6) <= writedata(7 downto 4);
						maze_desc(1)(7) <= writedata(3 downto 0);
					elsif address = "0000000000011100" then
						maze_desc(1)(8) <= writedata(15 downto 12);
						maze_desc(1)(9) <= writedata(11 downto 8);
						maze_desc(1)(10) <= writedata(7 downto 4);
						maze_desc(1)(11) <= writedata(3 downto 0);
					elsif address = "0000000000011110" then
						maze_desc(1)(12) <= writedata(15 downto 12);
						maze_desc(1)(13) <= writedata(11 downto 8);
						maze_desc(1)(14) <= writedata(7 downto 4);
						maze_desc(1)(15) <= writedata(3 downto 0);
						
					-- third row of maze
					elsif address = "0000000000100000" then
						maze_desc(2)(0) <= writedata(15 downto 12);
						maze_desc(2)(1) <= writedata(11 downto 8);
						maze_desc(2)(2) <= writedata(7 downto 4);
						maze_desc(2)(3) <= writedata(3 downto 0);
					elsif address = "0000000000100010" then
						maze_desc(2)(4) <= writedata(15 downto 12);
						maze_desc(2)(5) <= writedata(11 downto 8);
						maze_desc(2)(6) <= writedata(7 downto 4);
						maze_desc(2)(7) <= writedata(3 downto 0);
					elsif address = "0000000000100100" then
						maze_desc(2)(8) <= writedata(15 downto 12);
						maze_desc(2)(9) <= writedata(11 downto 8);
						maze_desc(2)(10) <= writedata(7 downto 4);
						maze_desc(2)(11) <= writedata(3 downto 0);
					elsif address = "0000000000100110" then
						maze_desc(2)(12) <= writedata(15 downto 12);
						maze_desc(2)(13) <= writedata(11 downto 8);
						maze_desc(2)(14) <= writedata(7 downto 4);
						maze_desc(2)(15) <= writedata(3 downto 0);
						
					-- fourth row of maze
					elsif address = "0000000000101000" then
						maze_desc(3)(0) <= writedata(15 downto 12);
						maze_desc(3)(1) <= writedata(11 downto 8);
						maze_desc(3)(2) <= writedata(7 downto 4);
						maze_desc(3)(3) <= writedata(3 downto 0);
					elsif address = "0000000000101010" then
						maze_desc(3)(4) <= writedata(15 downto 12);
						maze_desc(3)(5) <= writedata(11 downto 8);
						maze_desc(3)(6) <= writedata(7 downto 4);
						maze_desc(3)(7) <= writedata(3 downto 0);
					elsif address = "0000000000101100" then
						maze_desc(3)(8) <= writedata(15 downto 12);
						maze_desc(3)(9) <= writedata(11 downto 8);
						maze_desc(3)(10) <= writedata(7 downto 4);
						maze_desc(3)(11) <= writedata(3 downto 0);
					elsif address = "0000000000101110" then
						maze_desc(3)(12) <= writedata(15 downto 12);
						maze_desc(3)(13) <= writedata(11 downto 8);
						maze_desc(3)(14) <= writedata(7 downto 4);
						maze_desc(3)(15) <= writedata(3 downto 0);	
					
					-- fifth row of maze
					elsif address = "0000000000110000" then
						maze_desc(4)(0) <= writedata(15 downto 12);
						maze_desc(4)(1) <= writedata(11 downto 8);
						maze_desc(4)(2) <= writedata(7 downto 4);
						maze_desc(4)(3) <= writedata(3 downto 0);
					elsif address = "0000000000110010" then
						maze_desc(4)(4) <= writedata(15 downto 12);
						maze_desc(4)(5) <= writedata(11 downto 8);
						maze_desc(4)(6) <= writedata(7 downto 4);
						maze_desc(4)(7) <= writedata(3 downto 0);
					elsif address = "0000000000110100" then
						maze_desc(4)(8) <= writedata(15 downto 12);
						maze_desc(4)(9) <= writedata(11 downto 8);
						maze_desc(4)(10) <= writedata(7 downto 4);
						maze_desc(4)(11) <= writedata(3 downto 0);
					elsif address = "0000000000110110" then
						maze_desc(4)(12) <= writedata(15 downto 12);
						maze_desc(4)(13) <= writedata(11 downto 8);
						maze_desc(4)(14) <= writedata(7 downto 4);
						maze_desc(4)(15) <= writedata(3 downto 0);	
					
					-- sixth row of maze
					elsif address = "0000000000111000" then
						maze_desc(5)(0) <= writedata(15 downto 12);
						maze_desc(5)(1) <= writedata(11 downto 8);
						maze_desc(5)(2) <= writedata(7 downto 4);
						maze_desc(5)(3) <= writedata(3 downto 0);
					elsif address = "0000000000111010" then
						maze_desc(5)(4) <= writedata(15 downto 12);
						maze_desc(5)(5) <= writedata(11 downto 8);
						maze_desc(5)(6) <= writedata(7 downto 4);
						maze_desc(5)(7) <= writedata(3 downto 0);
					elsif address = "0000000000111100" then
						maze_desc(5)(8) <= writedata(15 downto 12);
						maze_desc(5)(9) <= writedata(11 downto 8);
						maze_desc(5)(10) <= writedata(7 downto 4);
						maze_desc(5)(11) <= writedata(3 downto 0);
					elsif address = "0000000000111110" then
						maze_desc(5)(12) <= writedata(15 downto 12);
						maze_desc(5)(13) <= writedata(11 downto 8);
						maze_desc(5)(14) <= writedata(7 downto 4);
						maze_desc(5)(15) <= writedata(3 downto 0);	
						
					-- seventh row of maze
					elsif address = "0000000001000000" then
						maze_desc(6)(0) <= writedata(15 downto 12);
						maze_desc(6)(1) <= writedata(11 downto 8);
						maze_desc(6)(2) <= writedata(7 downto 4);
						maze_desc(6)(3) <= writedata(3 downto 0);
					elsif address = "0000000001000010" then
						maze_desc(6)(4) <= writedata(15 downto 12);
						maze_desc(6)(5) <= writedata(11 downto 8);
						maze_desc(6)(6) <= writedata(7 downto 4);
						maze_desc(6)(7) <= writedata(3 downto 0);
					elsif address = "0000000001000100" then
						maze_desc(6)(8) <= writedata(15 downto 12);
						maze_desc(6)(9) <= writedata(11 downto 8);
						maze_desc(6)(10) <= writedata(7 downto 4);
						maze_desc(6)(11) <= writedata(3 downto 0);
					elsif address = "0000000001000110" then
						maze_desc(6)(12) <= writedata(15 downto 12);
						maze_desc(6)(13) <= writedata(11 downto 8);
						maze_desc(6)(14) <= writedata(7 downto 4);
						maze_desc(6)(15) <= writedata(3 downto 0);
					
					-- eigth row of maze
					elsif address = "0000000001001000" then
						maze_desc(7)(0) <= writedata(15 downto 12);
						maze_desc(7)(1) <= writedata(11 downto 8);
						maze_desc(7)(2) <= writedata(7 downto 4);
						maze_desc(7)(3) <= writedata(3 downto 0);
					elsif address = "0000000001001010" then
						maze_desc(7)(4) <= writedata(15 downto 12);
						maze_desc(7)(5) <= writedata(11 downto 8);
						maze_desc(7)(6) <= writedata(7 downto 4);
						maze_desc(7)(7) <= writedata(3 downto 0);
					elsif address = "0000000001001100" then
						maze_desc(7)(8) <= writedata(15 downto 12);
						maze_desc(7)(9) <= writedata(11 downto 8);
						maze_desc(7)(10) <= writedata(7 downto 4);
						maze_desc(7)(11) <= writedata(3 downto 0);
					elsif address = "0000000001001110" then
						maze_desc(7)(12) <= writedata(15 downto 12);
						maze_desc(7)(13) <= writedata(11 downto 8);
						maze_desc(7)(14) <= writedata(7 downto 4);
						maze_desc(7)(15) <= writedata(3 downto 0);
					
					-- ninth row of maze
					elsif address = "0000000001010000" then
						maze_desc(8)(0) <= writedata(15 downto 12);
						maze_desc(8)(1) <= writedata(11 downto 8);
						maze_desc(8)(2) <= writedata(7 downto 4);
						maze_desc(8)(3) <= writedata(3 downto 0);
					elsif address = "0000000001010010" then
						maze_desc(8)(4) <= writedata(15 downto 12);
						maze_desc(8)(5) <= writedata(11 downto 8);
						maze_desc(8)(6) <= writedata(7 downto 4);
						maze_desc(8)(7) <= writedata(3 downto 0);
					elsif address = "0000000001010100" then
						maze_desc(8)(8) <= writedata(15 downto 12);
						maze_desc(8)(9) <= writedata(11 downto 8);
						maze_desc(8)(10) <= writedata(7 downto 4);
						maze_desc(8)(11) <= writedata(3 downto 0);
					elsif address = "0000000001010110" then
						maze_desc(8)(12) <= writedata(15 downto 12);
						maze_desc(8)(13) <= writedata(11 downto 8);
						maze_desc(8)(14) <= writedata(7 downto 4);
						maze_desc(8)(15) <= writedata(3 downto 0);
						
					-- tenth row of maze
					elsif address = "0000000001011000" then
						maze_desc(9)(0) <= writedata(15 downto 12);
						maze_desc(9)(1) <= writedata(11 downto 8);
						maze_desc(9)(2) <= writedata(7 downto 4);
						maze_desc(9)(3) <= writedata(3 downto 0);
					elsif address = "0000000001011010" then
						maze_desc(9)(4) <= writedata(15 downto 12);
						maze_desc(9)(5) <= writedata(11 downto 8);
						maze_desc(9)(6) <= writedata(7 downto 4);
						maze_desc(9)(7) <= writedata(3 downto 0);
					elsif address = "0000000001011100" then
						maze_desc(9)(8) <= writedata(15 downto 12);
						maze_desc(9)(9) <= writedata(11 downto 8);
						maze_desc(9)(10) <= writedata(7 downto 4);
						maze_desc(9)(11) <= writedata(3 downto 0);
					elsif address = "0000000001011110" then
						maze_desc(9)(12) <= writedata(15 downto 12);
						maze_desc(9)(13) <= writedata(11 downto 8);
						maze_desc(9)(14) <= writedata(7 downto 4);
						maze_desc(9)(15) <= writedata(3 downto 0);
						
					-- eleventh row of maze
					elsif address = "0000000001100000" then
						maze_desc(10)(0) <= writedata(15 downto 12);
						maze_desc(10)(1) <= writedata(11 downto 8);
						maze_desc(10)(2) <= writedata(7 downto 4);
						maze_desc(10)(3) <= writedata(3 downto 0);
					elsif address = "0000000001100010" then
						maze_desc(10)(4) <= writedata(15 downto 12);
						maze_desc(10)(5) <= writedata(11 downto 8);
						maze_desc(10)(6) <= writedata(7 downto 4);
						maze_desc(10)(7) <= writedata(3 downto 0);
					elsif address = "0000000001100100" then
						maze_desc(10)(8) <= writedata(15 downto 12);
						maze_desc(10)(9) <= writedata(11 downto 8);
						maze_desc(10)(10) <= writedata(7 downto 4);
						maze_desc(10)(11) <= writedata(3 downto 0);
					elsif address = "0000000001100110" then
						maze_desc(10)(12) <= writedata(15 downto 12);
						maze_desc(10)(13) <= writedata(11 downto 8);
						maze_desc(10)(14) <= writedata(7 downto 4);
						maze_desc(10)(15) <= writedata(3 downto 0);
						
					-- twelfth row of maze
					elsif address = "0000000001101000" then
						maze_desc(11)(0) <= writedata(15 downto 12);
						maze_desc(11)(1) <= writedata(11 downto 8);
						maze_desc(11)(2) <= writedata(7 downto 4);
						maze_desc(11)(3) <= writedata(3 downto 0);
					elsif address = "0000000001101010" then
						maze_desc(11)(4) <= writedata(15 downto 12);
						maze_desc(11)(5) <= writedata(11 downto 8);
						maze_desc(11)(6) <= writedata(7 downto 4);
						maze_desc(11)(7) <= writedata(3 downto 0);
					elsif address = "0000000001101100" then
						maze_desc(11)(8) <= writedata(15 downto 12);
						maze_desc(11)(9) <= writedata(11 downto 8);
						maze_desc(11)(10) <= writedata(7 downto 4);
						maze_desc(11)(11) <= writedata(3 downto 0);
					elsif address = "0000000001101110" then
						maze_desc(11)(12) <= writedata(15 downto 12);
						maze_desc(11)(13) <= writedata(11 downto 8);
						maze_desc(11)(14) <= writedata(7 downto 4);
						maze_desc(11)(15) <= writedata(3 downto 0);		
						
	--				 thirteenth row of maze
					elsif address = "0000000001110000" then
						maze_desc(12)(0) <= writedata(15 downto 12);
						maze_desc(12)(1) <= writedata(11 downto 8);
						maze_desc(12)(2) <= writedata(7 downto 4);
						maze_desc(12)(3) <= writedata(3 downto 0);
					elsif address = "0000000001110010" then
						maze_desc(12)(4) <= writedata(15 downto 12);
						maze_desc(12)(5) <= writedata(11 downto 8);
						maze_desc(12)(6) <= writedata(7 downto 4);
						maze_desc(12)(7) <= writedata(3 downto 0);
					elsif address = "0000000001110100" then
						maze_desc(12)(8) <= writedata(15 downto 12);
						maze_desc(12)(9) <= writedata(11 downto 8);
						maze_desc(12)(10) <= writedata(7 downto 4);
						maze_desc(12)(11) <= writedata(3 downto 0);
					elsif address = "0000000001110110" then
						maze_desc(12)(12) <= writedata(15 downto 12);
						maze_desc(12)(13) <= writedata(11 downto 8);
						maze_desc(12)(14) <= writedata(7 downto 4);
						maze_desc(12)(15) <= writedata(3 downto 0);			
						
	--				 fourteenth row of maze
					elsif address = "0000000001111000" then
						maze_desc(13)(0) <= writedata(15 downto 12);
						maze_desc(13)(1) <= writedata(11 downto 8);
						maze_desc(13)(2) <= writedata(7 downto 4);
						maze_desc(13)(3) <= writedata(3 downto 0);
					elsif address = "0000000001111010" then
						maze_desc(13)(4) <= writedata(15 downto 12);
						maze_desc(13)(5) <= writedata(11 downto 8);
						maze_desc(13)(6) <= writedata(7 downto 4);
						maze_desc(13)(7) <= writedata(3 downto 0);
					elsif address = "0000000001111100" then
						maze_desc(13)(8) <= writedata(15 downto 12);
						maze_desc(13)(9) <= writedata(11 downto 8);
						maze_desc(13)(10) <= writedata(7 downto 4);
						maze_desc(13)(11) <= writedata(3 downto 0);
					elsif address = "0000000001111110" then
						maze_desc(13)(12) <= writedata(15 downto 12);
						maze_desc(13)(13) <= writedata(11 downto 8);
						maze_desc(13)(14) <= writedata(7 downto 4);
						maze_desc(13)(15) <= writedata(3 downto 0);		
						
	--				 ffiteenth row of maze
					elsif address = "0000000010000000" then
						maze_desc(14)(0) <= writedata(15 downto 12);
						maze_desc(14)(1) <= writedata(11 downto 8);
						maze_desc(14)(2) <= writedata(7 downto 4);
						maze_desc(14)(3) <= writedata(3 downto 0);
					elsif address = "0000000010000010" then
						maze_desc(14)(4) <= writedata(15 downto 12);
						maze_desc(14)(5) <= writedata(11 downto 8);
						maze_desc(14)(6) <= writedata(7 downto 4);
						maze_desc(14)(7) <= writedata(3 downto 0);
					elsif address = "0000000010000100" then
						maze_desc(14)(8) <= writedata(15 downto 12);
						maze_desc(14)(9) <= writedata(11 downto 8);
						maze_desc(14)(10) <= writedata(7 downto 4);
						maze_desc(14)(11) <= writedata(3 downto 0);
					elsif address = "0000000010000110" then
						maze_desc(14)(12) <= writedata(15 downto 12);
						maze_desc(14)(13) <= writedata(11 downto 8);
						maze_desc(14)(14) <= writedata(7 downto 4);
						maze_desc(14)(15) <= writedata(3 downto 0);		
						
	--				sixteenth row of maze
					elsif address = "0000000010001000" then
						maze_desc(15)(0) <= writedata(15 downto 12);
						maze_desc(15)(1) <= writedata(11 downto 8);
						maze_desc(15)(2) <= writedata(7 downto 4);
						maze_desc(15)(3) <= writedata(3 downto 0);
					elsif address = "0000000010001010" then
						maze_desc(15)(4) <= writedata(15 downto 12);
						maze_desc(15)(5) <= writedata(11 downto 8);
						maze_desc(15)(6) <= writedata(7 downto 4);
						maze_desc(15)(7) <= writedata(3 downto 0);
					elsif address = "0000000010001100" then
						maze_desc(15)(8) <= writedata(15 downto 12);
						maze_desc(15)(9) <= writedata(11 downto 8);
						maze_desc(15)(10) <= writedata(7 downto 4);
						maze_desc(15)(11) <= writedata(3 downto 0);
					elsif address = "0000000010001110" then
						maze_desc(15)(12) <= writedata(15 downto 12);
						maze_desc(15)(13) <= writedata(11 downto 8);
						maze_desc(15)(14) <= writedata(7 downto 4);
						maze_desc(15)(15) <= writedata(3 downto 0);			
						
				
					-- top row of obj
					elsif address = "0000000010010010" then
						obj_desc(0)(0) <= writedata(15 downto 12);
						obj_desc(0)(1) <= writedata(11 downto 8);
						obj_desc(0)(2) <= writedata(7 downto 4);
						obj_desc(0)(3) <= writedata(3 downto 0);
					elsif address = "0000000010010100" then
						obj_desc(0)(4) <= writedata(15 downto 12);
						obj_desc(0)(5) <= writedata(11 downto 8);
						obj_desc(0)(6) <= writedata(7 downto 4);
						obj_desc(0)(7) <= writedata(3 downto 0);
					elsif address = "0000000010010110" then
						obj_desc(0)(8) <= writedata(15 downto 12);
						obj_desc(0)(9) <= writedata(11 downto 8);
						obj_desc(0)(10) <= writedata(7 downto 4);
						obj_desc(0)(11) <= writedata(3 downto 0);
					elsif address = "0000000010011000" then
						obj_desc(0)(12) <= writedata(15 downto 12);
						obj_desc(0)(13) <= writedata(11 downto 8);
						obj_desc(0)(14) <= writedata(7 downto 4);
						obj_desc(0)(15) <= writedata(3 downto 0);
					
					-- second row of obj
					elsif address = "0000000010011010" then
						obj_desc(1)(0) <= writedata(15 downto 12);
						obj_desc(1)(1) <= writedata(11 downto 8);
						obj_desc(1)(2) <= writedata(7 downto 4);
						obj_desc(1)(3) <= writedata(3 downto 0);
					elsif address = "0000000010011100" then
						obj_desc(1)(4) <= writedata(15 downto 12);
						obj_desc(1)(5) <= writedata(11 downto 8);
						obj_desc(1)(6) <= writedata(7 downto 4);
						obj_desc(1)(7) <= writedata(3 downto 0);
					elsif address = "0000000010011110" then
						obj_desc(1)(8) <= writedata(15 downto 12);
						obj_desc(1)(9) <= writedata(11 downto 8);
						obj_desc(1)(10) <= writedata(7 downto 4);
						obj_desc(1)(11) <= writedata(3 downto 0);
					elsif address = "0000000010100000" then
						obj_desc(1)(12) <= writedata(15 downto 12);
						obj_desc(1)(13) <= writedata(11 downto 8);
						obj_desc(1)(14) <= writedata(7 downto 4);
						obj_desc(1)(15) <= writedata(3 downto 0);
						
					-- third row of obj
					elsif address = "0000000010100010" then
						obj_desc(2)(0) <= writedata(15 downto 12);
						obj_desc(2)(1) <= writedata(11 downto 8);
						obj_desc(2)(2) <= writedata(7 downto 4);
						obj_desc(2)(3) <= writedata(3 downto 0);
					elsif address = "0000000010100100" then
						obj_desc(2)(4) <= writedata(15 downto 12);
						obj_desc(2)(5) <= writedata(11 downto 8);
						obj_desc(2)(6) <= writedata(7 downto 4);
						obj_desc(2)(7) <= writedata(3 downto 0);
					elsif address = "0000000010100110" then
						obj_desc(2)(8) <= writedata(15 downto 12);
						obj_desc(2)(9) <= writedata(11 downto 8);
						obj_desc(2)(10) <= writedata(7 downto 4);
						obj_desc(2)(11) <= writedata(3 downto 0);
					elsif address = "0000000010101000" then
						obj_desc(2)(12) <= writedata(15 downto 12);
						obj_desc(2)(13) <= writedata(11 downto 8);
						obj_desc(2)(14) <= writedata(7 downto 4);
						obj_desc(2)(15) <= writedata(3 downto 0);
						
					-- fourth row of obj
					elsif address = "0000000010101010" then
						obj_desc(3)(0) <= writedata(15 downto 12);
						obj_desc(3)(1) <= writedata(11 downto 8);
						obj_desc(3)(2) <= writedata(7 downto 4);
						obj_desc(3)(3) <= writedata(3 downto 0);
					elsif address = "0000000010101100" then
						obj_desc(3)(4) <= writedata(15 downto 12);
						obj_desc(3)(5) <= writedata(11 downto 8);
						obj_desc(3)(6) <= writedata(7 downto 4);
						obj_desc(3)(7) <= writedata(3 downto 0);
					elsif address = "0000000010101110" then
						obj_desc(3)(8) <= writedata(15 downto 12);
						obj_desc(3)(9) <= writedata(11 downto 8);
						obj_desc(3)(10) <= writedata(7 downto 4);
						obj_desc(3)(11) <= writedata(3 downto 0);
					elsif address = "0000000010110000" then
						obj_desc(3)(12) <= writedata(15 downto 12);
						obj_desc(3)(13) <= writedata(11 downto 8);
						obj_desc(3)(14) <= writedata(7 downto 4);
						obj_desc(3)(15) <= writedata(3 downto 0);	
					
					-- fifth row of obj
					elsif address = "0000000010110010" then
						obj_desc(4)(0) <= writedata(15 downto 12);
						obj_desc(4)(1) <= writedata(11 downto 8);
						obj_desc(4)(2) <= writedata(7 downto 4);
						obj_desc(4)(3) <= writedata(3 downto 0);
					elsif address = "0000000010110100" then
						obj_desc(4)(4) <= writedata(15 downto 12);
						obj_desc(4)(5) <= writedata(11 downto 8);
						obj_desc(4)(6) <= writedata(7 downto 4);
						obj_desc(4)(7) <= writedata(3 downto 0);
					elsif address = "0000000010110110" then
						obj_desc(4)(8) <= writedata(15 downto 12);
						obj_desc(4)(9) <= writedata(11 downto 8);
						obj_desc(4)(10) <= writedata(7 downto 4);
						obj_desc(4)(11) <= writedata(3 downto 0);
					elsif address = "0000000010111000" then
						obj_desc(4)(12) <= writedata(15 downto 12);
						obj_desc(4)(13) <= writedata(11 downto 8);
						obj_desc(4)(14) <= writedata(7 downto 4);
						obj_desc(4)(15) <= writedata(3 downto 0);	
					
					-- sixth row of obj
					elsif address = "0000000010111010" then
						obj_desc(5)(0) <= writedata(15 downto 12);
						obj_desc(5)(1) <= writedata(11 downto 8);
						obj_desc(5)(2) <= writedata(7 downto 4);
						obj_desc(5)(3) <= writedata(3 downto 0);
					elsif address = "0000000010111100" then
						obj_desc(5)(4) <= writedata(15 downto 12);
						obj_desc(5)(5) <= writedata(11 downto 8);
						obj_desc(5)(6) <= writedata(7 downto 4);
						obj_desc(5)(7) <= writedata(3 downto 0);
					elsif address = "0000000010111110" then
						obj_desc(5)(8) <= writedata(15 downto 12);
						obj_desc(5)(9) <= writedata(11 downto 8);
						obj_desc(5)(10) <= writedata(7 downto 4);
						obj_desc(5)(11) <= writedata(3 downto 0);
					elsif address = "0000000011000000" then
						obj_desc(5)(12) <= writedata(15 downto 12);
						obj_desc(5)(13) <= writedata(11 downto 8);
						obj_desc(5)(14) <= writedata(7 downto 4);
						obj_desc(5)(15) <= writedata(3 downto 0);	
						
					-- seventh row of obj
					elsif address = "0000000011000010" then
						obj_desc(6)(0) <= writedata(15 downto 12);
						obj_desc(6)(1) <= writedata(11 downto 8);
						obj_desc(6)(2) <= writedata(7 downto 4);
						obj_desc(6)(3) <= writedata(3 downto 0);
					elsif address = "0000000011000100" then
						obj_desc(6)(4) <= writedata(15 downto 12);
						obj_desc(6)(5) <= writedata(11 downto 8);
						obj_desc(6)(6) <= writedata(7 downto 4);
						obj_desc(6)(7) <= writedata(3 downto 0);
					elsif address = "0000000011000110" then
						obj_desc(6)(8) <= writedata(15 downto 12);
						obj_desc(6)(9) <= writedata(11 downto 8);
						obj_desc(6)(10) <= writedata(7 downto 4);
						obj_desc(6)(11) <= writedata(3 downto 0);
					elsif address = "0000000011001000" then
						obj_desc(6)(12) <= writedata(15 downto 12);
						obj_desc(6)(13) <= writedata(11 downto 8);
						obj_desc(6)(14) <= writedata(7 downto 4);
						obj_desc(6)(15) <= writedata(3 downto 0);
					
					-- eigth row of obj
					elsif address = "0000000011001010" then
						obj_desc(7)(0) <= writedata(15 downto 12);
						obj_desc(7)(1) <= writedata(11 downto 8);
						obj_desc(7)(2) <= writedata(7 downto 4);
						obj_desc(7)(3) <= writedata(3 downto 0);
					elsif address = "0000000011001100" then
						obj_desc(7)(4) <= writedata(15 downto 12);
						obj_desc(7)(5) <= writedata(11 downto 8);
						obj_desc(7)(6) <= writedata(7 downto 4);
						obj_desc(7)(7) <= writedata(3 downto 0);
					elsif address = "0000000011001110" then
						obj_desc(7)(8) <= writedata(15 downto 12);
						obj_desc(7)(9) <= writedata(11 downto 8);
						obj_desc(7)(10) <= writedata(7 downto 4);
						obj_desc(7)(11) <= writedata(3 downto 0);
					elsif address = "0000000011010000" then
						obj_desc(7)(12) <= writedata(15 downto 12);
						obj_desc(7)(13) <= writedata(11 downto 8);
						obj_desc(7)(14) <= writedata(7 downto 4);
						obj_desc(7)(15) <= writedata(3 downto 0);
					
					-- ninth row of obj
					elsif address = "0000000011010010" then
						obj_desc(8)(0) <= writedata(15 downto 12);
						obj_desc(8)(1) <= writedata(11 downto 8);
						obj_desc(8)(2) <= writedata(7 downto 4);
						obj_desc(8)(3) <= writedata(3 downto 0);
					elsif address = "0000000011010100" then
						obj_desc(8)(4) <= writedata(15 downto 12);
						obj_desc(8)(5) <= writedata(11 downto 8);
						obj_desc(8)(6) <= writedata(7 downto 4);
						obj_desc(8)(7) <= writedata(3 downto 0);
					elsif address = "0000000011010110" then
						obj_desc(8)(8) <= writedata(15 downto 12);
						obj_desc(8)(9) <= writedata(11 downto 8);
						obj_desc(8)(10) <= writedata(7 downto 4);
						obj_desc(8)(11) <= writedata(3 downto 0);
					elsif address = "0000000011011000" then
						obj_desc(8)(12) <= writedata(15 downto 12);
						obj_desc(8)(13) <= writedata(11 downto 8);
						obj_desc(8)(14) <= writedata(7 downto 4);
						obj_desc(8)(15) <= writedata(3 downto 0);
						
					-- tenth row of obj
					elsif address = "0000000011011010" then
						obj_desc(9)(0) <= writedata(15 downto 12);
						obj_desc(9)(1) <= writedata(11 downto 8);
						obj_desc(9)(2) <= writedata(7 downto 4);
						obj_desc(9)(3) <= writedata(3 downto 0);
					elsif address = "0000000011011100" then
						obj_desc(9)(4) <= writedata(15 downto 12);
						obj_desc(9)(5) <= writedata(11 downto 8);
						obj_desc(9)(6) <= writedata(7 downto 4);
						obj_desc(9)(7) <= writedata(3 downto 0);
					elsif address = "0000000011011110" then
						obj_desc(9)(8) <= writedata(15 downto 12);
						obj_desc(9)(9) <= writedata(11 downto 8);
						obj_desc(9)(10) <= writedata(7 downto 4);
						obj_desc(9)(11) <= writedata(3 downto 0);
					elsif address = "0000000011100000" then
						obj_desc(9)(12) <= writedata(15 downto 12);
						obj_desc(9)(13) <= writedata(11 downto 8);
						obj_desc(9)(14) <= writedata(7 downto 4);
						obj_desc(9)(15) <= writedata(3 downto 0);
						
					-- eleventh row of obj
					elsif address = "0000000011100010" then
						obj_desc(10)(0) <= writedata(15 downto 12);
						obj_desc(10)(1) <= writedata(11 downto 8);
						obj_desc(10)(2) <= writedata(7 downto 4);
						obj_desc(10)(3) <= writedata(3 downto 0);
					elsif address = "0000000011100100" then
						obj_desc(10)(4) <= writedata(15 downto 12);
						obj_desc(10)(5) <= writedata(11 downto 8);
						obj_desc(10)(6) <= writedata(7 downto 4);
						obj_desc(10)(7) <= writedata(3 downto 0);
					elsif address = "0000000011100110" then
						obj_desc(10)(8) <= writedata(15 downto 12);
						obj_desc(10)(9) <= writedata(11 downto 8);
						obj_desc(10)(10) <= writedata(7 downto 4);
						obj_desc(10)(11) <= writedata(3 downto 0);
					elsif address = "0000000011101000" then
						obj_desc(10)(12) <= writedata(15 downto 12);
						obj_desc(10)(13) <= writedata(11 downto 8);
						obj_desc(10)(14) <= writedata(7 downto 4);
						obj_desc(10)(15) <= writedata(3 downto 0);
						
					-- twelfth row of obj
					elsif address = "0000000011101010" then
						obj_desc(11)(0) <= writedata(15 downto 12);
						obj_desc(11)(1) <= writedata(11 downto 8);
						obj_desc(11)(2) <= writedata(7 downto 4);
						obj_desc(11)(3) <= writedata(3 downto 0);
					elsif address = "0000000011101100" then
						obj_desc(11)(4) <= writedata(15 downto 12);
						obj_desc(11)(5) <= writedata(11 downto 8);
						obj_desc(11)(6) <= writedata(7 downto 4);
						obj_desc(11)(7) <= writedata(3 downto 0);
					elsif address = "0000000011101110" then
						obj_desc(11)(8) <= writedata(15 downto 12);
						obj_desc(11)(9) <= writedata(11 downto 8);
						obj_desc(11)(10) <= writedata(7 downto 4);
						obj_desc(11)(11) <= writedata(3 downto 0);
					elsif address = "0000000011110000" then
						obj_desc(11)(12) <= writedata(15 downto 12);
						obj_desc(11)(13) <= writedata(11 downto 8);
						obj_desc(11)(14) <= writedata(7 downto 4);
						obj_desc(11)(15) <= writedata(3 downto 0);		
						
	--				 thirteenth row of obj
					elsif address = "0000000011110010" then
						obj_desc(12)(0) <= writedata(15 downto 12);
						obj_desc(12)(1) <= writedata(11 downto 8);
						obj_desc(12)(2) <= writedata(7 downto 4);
						obj_desc(12)(3) <= writedata(3 downto 0);
					elsif address = "0000000011110100" then
						obj_desc(12)(4) <= writedata(15 downto 12);
						obj_desc(12)(5) <= writedata(11 downto 8);
						obj_desc(12)(6) <= writedata(7 downto 4);
						obj_desc(12)(7) <= writedata(3 downto 0);
					elsif address = "0000000011110110" then
						obj_desc(12)(8) <= writedata(15 downto 12);
						obj_desc(12)(9) <= writedata(11 downto 8);
						obj_desc(12)(10) <= writedata(7 downto 4);
						obj_desc(12)(11) <= writedata(3 downto 0);
					elsif address = "0000000011111000" then
						obj_desc(12)(12) <= writedata(15 downto 12);
						obj_desc(12)(13) <= writedata(11 downto 8);
						obj_desc(12)(14) <= writedata(7 downto 4);
						obj_desc(12)(15) <= writedata(3 downto 0);			
						
	--				 fourteenth row of obj
					elsif address = "0000000011111010" then
						obj_desc(13)(0) <= writedata(15 downto 12);
						obj_desc(13)(1) <= writedata(11 downto 8);
						obj_desc(13)(2) <= writedata(7 downto 4);
						obj_desc(13)(3) <= writedata(3 downto 0);
					elsif address = "0000000011111100" then
						obj_desc(13)(4) <= writedata(15 downto 12);
						obj_desc(13)(5) <= writedata(11 downto 8);
						obj_desc(13)(6) <= writedata(7 downto 4);
						obj_desc(13)(7) <= writedata(3 downto 0);
					elsif address = "0000000011111110" then
						obj_desc(13)(8) <= writedata(15 downto 12);
						obj_desc(13)(9) <= writedata(11 downto 8);
						obj_desc(13)(10) <= writedata(7 downto 4);
						obj_desc(13)(11) <= writedata(3 downto 0);
					elsif address = "0000000100000000" then
						obj_desc(13)(12) <= writedata(15 downto 12);
						obj_desc(13)(13) <= writedata(11 downto 8);
						obj_desc(13)(14) <= writedata(7 downto 4);
						obj_desc(13)(15) <= writedata(3 downto 0);		
						
	--				 fifteenth row of obj
					elsif address = "0000000100000010" then
						obj_desc(14)(0) <= writedata(15 downto 12);
						obj_desc(14)(1) <= writedata(11 downto 8);
						obj_desc(14)(2) <= writedata(7 downto 4);
						obj_desc(14)(3) <= writedata(3 downto 0);
					elsif address = "0000000100000100" then
						obj_desc(14)(4) <= writedata(15 downto 12);
						obj_desc(14)(5) <= writedata(11 downto 8);
						obj_desc(14)(6) <= writedata(7 downto 4);
						obj_desc(14)(7) <= writedata(3 downto 0);
					elsif address = "0000000100000110" then
						obj_desc(14)(8) <= writedata(15 downto 12);
						obj_desc(14)(9) <= writedata(11 downto 8);
						obj_desc(14)(10) <= writedata(7 downto 4);
						obj_desc(14)(11) <= writedata(3 downto 0);
					elsif address = "0000000100001000" then
						obj_desc(14)(12) <= writedata(15 downto 12);
						obj_desc(14)(13) <= writedata(11 downto 8);
						obj_desc(14)(14) <= writedata(7 downto 4);
						obj_desc(14)(15) <= writedata(3 downto 0);		
						
	--				sixteenth row of obj
					elsif address = "0000000100001010" then
						obj_desc(15)(0) <= writedata(15 downto 12);
						obj_desc(15)(1) <= writedata(11 downto 8);
						obj_desc(15)(2) <= writedata(7 downto 4);
						obj_desc(15)(3) <= writedata(3 downto 0);
					elsif address = "0000000100001100" then
						obj_desc(15)(4) <= writedata(15 downto 12);
						obj_desc(15)(5) <= writedata(11 downto 8);
						obj_desc(15)(6) <= writedata(7 downto 4);
						obj_desc(15)(7) <= writedata(3 downto 0);
					elsif address = "0000000100001110" then
						obj_desc(15)(8) <= writedata(15 downto 12);
						obj_desc(15)(9) <= writedata(11 downto 8);
						obj_desc(15)(10) <= writedata(7 downto 4);
						obj_desc(15)(11) <= writedata(3 downto 0);
					elsif address = "0000000100010000" then
						obj_desc(15)(12) <= writedata(15 downto 12);
						obj_desc(15)(13) <= writedata(11 downto 8);
						obj_desc(15)(14) <= writedata(7 downto 4);
						obj_desc(15)(15) <= writedata(3 downto 0);				
					elsif address = "0000000100010010" then
						tmp_char_tile <= writedata;
					else 
						foo <= writedata;
					end if;
				elsif read = '1' then		
					if address = "0000000000000000" then
						readdata <= maze_x_offset;
					elsif address = "0000000000000010" then
						readdata <= maze_y_offset;
					elsif address = "0000000000000100" then
						readdata <= cur_xpos_pix;
					elsif address = "0000000000000110" then
						readdata <= cur_ypos_pix;
					elsif address = "0000000000001000" then
						readdata <= end_xpos_maze;	
					elsif address = "0000000000001010" then
						readdata <= end_ypos_maze;	
					elsif address = "0000000000001100" then
						readdata(15 downto 8) <= half_window_size_x;
						readdata(7 downto 0) <= half_window_size_y;
					elsif address = "0000000000001110" then
						readdata <= which_keys;
					elsif address = "0000000100010010" then
						readdata <= which_char_tile;
					
					elsif address = "0000000000010000" then
						readdata(15 downto 12) <= maze_desc(0)(0);
						readdata(11 downto 8) <= maze_desc(0)(1);
						readdata(7 downto 4) <= maze_desc(0)(2);
						readdata(3 downto 0) <= maze_desc(0)(3);
					elsif address = "0000000000010010" then
						readdata(15 downto 12) <= maze_desc(0)(4);
						readdata(11 downto 8) <= maze_desc(0)(5); 
						readdata(7 downto 4) <= maze_desc(0)(6); 
						readdata(3 downto 0) <= maze_desc(0)(7); 
					elsif address = "0000000000010100" then
						readdata(15 downto 12) <= maze_desc(0)(8); 
						readdata(11 downto 8) <= maze_desc(0)(9); 
						readdata(7 downto 4) <= maze_desc(0)(10); 
						readdata(3 downto 0) <= maze_desc(0)(11);
					elsif address = "0000000000010110" then
						readdata(15 downto 12) <= maze_desc(0)(12); 
						readdata(11 downto 8) <= maze_desc(0)(13); 
						readdata(7 downto 4) <= maze_desc(0)(14); 
						readdata(3 downto 0) <= maze_desc(0)(15); 
					
					-- second row of maze
					elsif address = "0000000000011000" then
						readdata(15 downto 12) <= maze_desc(1)(0); 
						readdata(11 downto 8) <= maze_desc(1)(1); 
						readdata(7 downto 4) <= maze_desc(1)(2); 
						readdata(3 downto 0) <= maze_desc(1)(3); 
					elsif address = "0000000000011010" then
						readdata(15 downto 12) <= maze_desc(1)(4); 
						readdata(11 downto 8) <= maze_desc(1)(5); 
						readdata(7 downto 4) <= maze_desc(1)(6); 
						readdata(3 downto 0) <= maze_desc(1)(7); 
					elsif address = "0000000000011100" then
						readdata(15 downto 12) <= maze_desc(1)(8);
						readdata(11 downto 8) <= maze_desc(1)(9); 
						readdata(7 downto 4) <= maze_desc(1)(10); 
						readdata(3 downto 0) <= maze_desc(1)(11); 
					elsif address = "0000000000011110" then
						readdata(15 downto 12) <= maze_desc(1)(12); 
						readdata(11 downto 8) <= maze_desc(1)(13); 
						readdata(7 downto 4) <= maze_desc(1)(14); 
						readdata(3 downto 0) <= maze_desc(1)(15); 
						
					-- third row of maze
					elsif address = "0000000000100000" then
						readdata(15 downto 12) <= maze_desc(2)(0); 
						readdata(11 downto 8) <= maze_desc(2)(1); 
						readdata(7 downto 4) <= maze_desc(2)(2); 
						readdata(3 downto 0) <= maze_desc(2)(3); 
					elsif address = "0000000000100010" then
						readdata(15 downto 12) <= maze_desc(2)(4); 
						readdata(11 downto 8) <= maze_desc(2)(5); 
						readdata(7 downto 4) <= maze_desc(2)(6); 
						readdata(3 downto 0) <= maze_desc(2)(7); 
					elsif address = "0000000000100100" then
						readdata(15 downto 12) <= maze_desc(2)(8); 
						readdata(11 downto 8) <= maze_desc(2)(9); 
						readdata(7 downto 4) <= maze_desc(2)(10); 
						readdata(3 downto 0) <= maze_desc(2)(11); 
					elsif address = "0000000000100110" then
						readdata(15 downto 12) <= maze_desc(2)(12); 
						readdata(11 downto 8) <= maze_desc(2)(13); 
						readdata(7 downto 4) <= maze_desc(2)(14); 
						readdata(3 downto 0) <= maze_desc(2)(15); 
						
					-- fourth row of maze
					elsif address = "0000000000101000" then
						readdata(15 downto 12) <= maze_desc(3)(0); 
						readdata(11 downto 8) <= maze_desc(3)(1); 
						readdata(7 downto 4) <= maze_desc(3)(2); 
						readdata(3 downto 0) <= maze_desc(3)(3); 
					elsif address = "0000000000101010" then
						readdata(15 downto 12) <= maze_desc(3)(4); 
						readdata(11 downto 8) <= maze_desc(3)(5); 
						readdata(7 downto 4) <= maze_desc(3)(6); 
						readdata(3 downto 0) <= maze_desc(3)(7); 
					elsif address = "0000000000101100" then
						readdata(15 downto 12) <= maze_desc(3)(8); 
						readdata(11 downto 8) <= maze_desc(3)(9); 
						readdata(7 downto 4) <= maze_desc(3)(10); 
						readdata(3 downto 0) <= maze_desc(3)(11); 
					elsif address = "0000000000101110" then
						readdata(15 downto 12) <= maze_desc(3)(12); 
						readdata(11 downto 8) <= maze_desc(3)(13); 
						readdata(7 downto 4) <= maze_desc(3)(14); 
						readdata(3 downto 0) <= maze_desc(3)(15); 	
					
					-- fifth row of maze
					elsif address = "0000000000110000" then
						readdata(15 downto 12) <= maze_desc(4)(0); 
						readdata(11 downto 8) <= maze_desc(4)(1); 
						readdata(7 downto 4) <= maze_desc(4)(2); 
						readdata(3 downto 0) <= maze_desc(4)(3); 
					elsif address = "0000000000110010" then
						readdata(15 downto 12) <= maze_desc(4)(4); 
						readdata(11 downto 8) <= maze_desc(4)(5); 
						readdata(7 downto 4) <= maze_desc(4)(6); 
						readdata(3 downto 0) <= maze_desc(4)(7); 
					elsif address = "0000000000110100" then
						readdata(15 downto 12) <= maze_desc(4)(8); 
						readdata(11 downto 8) <= maze_desc(4)(9); 
						readdata(7 downto 4) <= maze_desc(4)(10); 
						readdata(3 downto 0) <= maze_desc(4)(11); 
					elsif address = "0000000000110110" then
						readdata(15 downto 12) <= maze_desc(4)(12); 
						readdata(11 downto 8) <= maze_desc(4)(13); 
						readdata(7 downto 4) <= maze_desc(4)(14); 
						readdata(3 downto 0) <= maze_desc(4)(15); 	
					
					-- sixth row of maze
					elsif address = "0000000000111000" then
						readdata(15 downto 12) <= maze_desc(5)(0); 
						readdata(11 downto 8) <= maze_desc(5)(1); 
						readdata(7 downto 4) <= maze_desc(5)(2); 
						readdata(3 downto 0) <= maze_desc(5)(3); 
					elsif address = "0000000000111010" then
						readdata(15 downto 12) <= maze_desc(5)(4); 
						readdata(11 downto 8) <= maze_desc(5)(5); 
						readdata(7 downto 4) <= maze_desc(5)(6); 
						readdata(3 downto 0) <= maze_desc(5)(7); 
					elsif address = "0000000000111100" then
						readdata(15 downto 12) <= maze_desc(5)(8); 
						readdata(11 downto 8) <= maze_desc(5)(9); 
						readdata(7 downto 4) <= maze_desc(5)(10); 
						readdata(3 downto 0) <= maze_desc(5)(11); 
					elsif address = "0000000000111110" then
						readdata(15 downto 12) <= maze_desc(5)(12); 
						readdata(11 downto 8) <= maze_desc(5)(13); 
						readdata(7 downto 4) <= maze_desc(5)(14); 
						readdata(3 downto 0) <= maze_desc(5)(15); 	
						
					-- seventh row of maze
					elsif address = "0000000001000000" then
						readdata(15 downto 12) <= maze_desc(6)(0); 
						readdata(11 downto 8) <= maze_desc(6)(1); 
						readdata(7 downto 4) <= maze_desc(6)(2); 
						readdata(3 downto 0) <= maze_desc(6)(3); 
					elsif address = "0000000001000010" then
						readdata(15 downto 12) <= maze_desc(6)(4); 
						readdata(11 downto 8) <= maze_desc(6)(5); 
						readdata(7 downto 4) <= maze_desc(6)(6); 
						readdata(3 downto 0) <= maze_desc(6)(7); 
					elsif address = "0000000001000100" then
						readdata(15 downto 12) <= maze_desc(6)(8); 
						readdata(11 downto 8) <= maze_desc(6)(9); 
						readdata(7 downto 4) <= maze_desc(6)(10); 
						readdata(3 downto 0) <= maze_desc(6)(11); 
					elsif address = "0000000001000110" then
						readdata(15 downto 12) <= maze_desc(6)(12); 
						readdata(11 downto 8) <= maze_desc(6)(13); 
						readdata(7 downto 4) <= maze_desc(6)(14); 
						readdata(3 downto 0) <= maze_desc(6)(15); 
					
					-- eigth row of maze
					elsif address = "0000000001001000" then
						readdata(15 downto 12) <= maze_desc(7)(0); 
						readdata(11 downto 8) <= maze_desc(7)(1); 
						readdata(7 downto 4) <= maze_desc(7)(2); 
						readdata(3 downto 0) <= maze_desc(7)(3); 
					elsif address = "0000000001001010" then
						readdata(15 downto 12) <= maze_desc(7)(4); 
						readdata(11 downto 8) <= maze_desc(7)(5); 
						readdata(7 downto 4) <= maze_desc(7)(6); 
						readdata(3 downto 0) <= maze_desc(7)(7); 
					elsif address = "0000000001001100" then
						readdata(15 downto 12) <= maze_desc(7)(8); 
						readdata(11 downto 8) <= maze_desc(7)(9); 
						readdata(7 downto 4) <= maze_desc(7)(10); 
						readdata(3 downto 0) <= maze_desc(7)(11); 
					elsif address = "0000000001001110" then
						readdata(15 downto 12) <= maze_desc(7)(12); 
						readdata(11 downto 8) <= maze_desc(7)(13); 
						readdata(7 downto 4) <= maze_desc(7)(14); 
						readdata(3 downto 0) <= maze_desc(7)(15); 
					
					-- ninth row of maze
					elsif address = "0000000001010000" then
						readdata(15 downto 12) <= maze_desc(8)(0); 
						readdata(11 downto 8) <= maze_desc(8)(1); 
						readdata(7 downto 4) <= maze_desc(8)(2); 
						readdata(3 downto 0) <= maze_desc(8)(3); 
					elsif address = "0000000001010010" then
						readdata(15 downto 12) <= maze_desc(8)(4); 
						readdata(11 downto 8) <= maze_desc(8)(5); 
						readdata(7 downto 4) <= maze_desc(8)(6); 
						readdata(3 downto 0) <= maze_desc(8)(7); 
					elsif address = "0000000001010100" then
						readdata(15 downto 12) <= maze_desc(8)(8); 
						readdata(11 downto 8) <= maze_desc(8)(9); 
						readdata(7 downto 4) <= maze_desc(8)(10); 
						readdata(3 downto 0) <= maze_desc(8)(11); 
					elsif address = "0000000001010110" then
						readdata(15 downto 12) <= maze_desc(8)(12); 
						readdata(11 downto 8) <= maze_desc(8)(13); 
						readdata(7 downto 4) <= maze_desc(8)(14); 
						readdata(3 downto 0) <= maze_desc(8)(15); 
						
					-- tenth row of maze
					elsif address = "0000000001011000" then
						readdata(15 downto 12) <= maze_desc(9)(0); 
						readdata(11 downto 8) <= maze_desc(9)(1); 
						readdata(7 downto 4) <= maze_desc(9)(2); 
						readdata(3 downto 0) <= maze_desc(9)(3); 
					elsif address = "0000000001011010" then
						readdata(15 downto 12) <= maze_desc(9)(4); 
						readdata(11 downto 8) <= maze_desc(9)(5); 
						readdata(7 downto 4) <= maze_desc(9)(6); 
						readdata(3 downto 0) <= maze_desc(9)(7); 
					elsif address = "0000000001011100" then
						readdata(15 downto 12) <= maze_desc(9)(8); 
						readdata(11 downto 8) <= maze_desc(9)(9); 
						readdata(7 downto 4) <= maze_desc(9)(10); 
						readdata(3 downto 0) <= maze_desc(9)(11); 
					elsif address = "0000000001011110" then
						readdata(15 downto 12) <= maze_desc(9)(12); 
						readdata(11 downto 8) <= maze_desc(9)(13); 
						readdata(7 downto 4) <= maze_desc(9)(14); 
						readdata(3 downto 0) <= maze_desc(9)(15); 
						
					-- eleventh row of maze
					elsif address = "0000000001100000" then
						readdata(15 downto 12) <= maze_desc(10)(0); 
						readdata(11 downto 8) <= maze_desc(10)(1); 
						readdata(7 downto 4) <= maze_desc(10)(2); 
						readdata(3 downto 0) <= maze_desc(10)(3); 
					elsif address = "0000000001100010" then
						readdata(15 downto 12) <= maze_desc(10)(4); 
						readdata(11 downto 8) <= maze_desc(10)(5); 
						readdata(7 downto 4) <= maze_desc(10)(6); 
						readdata(3 downto 0) <= maze_desc(10)(7); 
					elsif address = "0000000001100100" then
						readdata(15 downto 12) <= maze_desc(10)(8); 
						readdata(11 downto 8) <= maze_desc(10)(9); 
						readdata(7 downto 4) <= maze_desc(10)(10); 
						readdata(3 downto 0) <= maze_desc(10)(11); 
					elsif address = "0000000001100110" then
						readdata(15 downto 12) <= maze_desc(10)(12); 
						readdata(11 downto 8) <= maze_desc(10)(13); 
						readdata(7 downto 4) <= maze_desc(10)(14); 
						readdata(3 downto 0) <= maze_desc(10)(15); 
						
					-- twelfth row of maze
					elsif address = "0000000001101000" then
						readdata(15 downto 12) <= maze_desc(11)(0); 
						readdata(11 downto 8) <= maze_desc(11)(1); 
						readdata(7 downto 4) <= maze_desc(11)(2); 
						readdata(3 downto 0) <= maze_desc(11)(3); 
					elsif address = "0000000001101010" then
						readdata(15 downto 12) <= maze_desc(11)(4); 
						readdata(11 downto 8) <= maze_desc(11)(5); 
						readdata(7 downto 4) <= maze_desc(11)(6); 
						readdata(3 downto 0) <= maze_desc(11)(7); 
					elsif address = "0000000001101100" then
						readdata(15 downto 12) <= maze_desc(11)(8); 
						readdata(11 downto 8) <= maze_desc(11)(9); 
						readdata(7 downto 4) <= maze_desc(11)(10); 
						readdata(3 downto 0) <= maze_desc(11)(11); 
					elsif address = "0000000001101110" then
						readdata(15 downto 12) <= maze_desc(11)(12); 
						readdata(11 downto 8) <= maze_desc(11)(13); 
						readdata(7 downto 4) <= maze_desc(11)(14); 
						readdata(3 downto 0) <= maze_desc(11)(15); 		
						
	--				 thirteenth row of maze
					elsif address = "0000000001110000" then
						readdata(15 downto 12) <= maze_desc(12)(0); 
						readdata(11 downto 8) <= maze_desc(12)(1); 
						readdata(7 downto 4) <= maze_desc(12)(2); 
						readdata(3 downto 0) <= maze_desc(12)(3); 
					elsif address = "0000000001110010" then
						readdata(15 downto 12) <= maze_desc(12)(4); 
						readdata(11 downto 8) <= maze_desc(12)(5); 
						readdata(7 downto 4) <= maze_desc(12)(6); 
						readdata(3 downto 0) <= maze_desc(12)(7); 
					elsif address = "0000000001110100" then
						readdata(15 downto 12) <= maze_desc(12)(8); 
						readdata(11 downto 8) <= maze_desc(12)(9); 
						readdata(7 downto 4) <= maze_desc(12)(10); 
						readdata(3 downto 0) <= maze_desc(12)(11); 
					elsif address = "0000000001110110" then
						readdata(15 downto 12) <= maze_desc(12)(12); 
						readdata(11 downto 8) <= maze_desc(12)(13); 
						readdata(7 downto 4) <= maze_desc(12)(14); 
						readdata(3 downto 0) <= maze_desc(12)(15); 			
						
	--				 fourteenth row of maze
					elsif address = "0000000001111000" then
						readdata(15 downto 12) <= maze_desc(13)(0); 
						readdata(11 downto 8) <= maze_desc(13)(1); 
						readdata(7 downto 4) <= maze_desc(13)(2); 
						readdata(3 downto 0) <= maze_desc(13)(3); 
					elsif address = "0000000001111010" then
						readdata(15 downto 12) <= maze_desc(13)(4); 
						readdata(11 downto 8) <= maze_desc(13)(5); 
						readdata(7 downto 4) <= maze_desc(13)(6); 
						readdata(3 downto 0) <= maze_desc(13)(7); 
					elsif address = "0000000001111100" then
						readdata(15 downto 12) <= maze_desc(13)(8); 
						readdata(11 downto 8) <= maze_desc(13)(9); 
						readdata(7 downto 4) <= maze_desc(13)(10); 
						readdata(3 downto 0) <= maze_desc(13)(11);
					elsif address = "0000000001111110" then
						readdata(15 downto 12) <= maze_desc(13)(12); 
						readdata(11 downto 8) <= maze_desc(13)(13); 
						readdata(7 downto 4) <= maze_desc(13)(14); 
						readdata(3 downto 0) <= maze_desc(13)(15); 		
						
	--				 ffiteenth row of maze
					elsif address = "0000000010000000" then
						readdata(15 downto 12) <= maze_desc(14)(0); 
						readdata(11 downto 8) <= maze_desc(14)(1); 
						readdata(7 downto 4) <= maze_desc(14)(2); 
						readdata(3 downto 0) <= maze_desc(14)(3); 
					elsif address = "0000000010000010" then
						readdata(15 downto 12) <= maze_desc(14)(4); 
						readdata(11 downto 8) <= maze_desc(14)(5); 
						readdata(7 downto 4) <= maze_desc(14)(6); 
						readdata(3 downto 0) <= maze_desc(14)(7); 
					elsif address = "0000000010000100" then
						readdata(15 downto 12) <= maze_desc(14)(8); 
						readdata(11 downto 8) <= maze_desc(14)(9); 
						readdata(7 downto 4) <= maze_desc(14)(10); 
						readdata(3 downto 0) <= maze_desc(14)(11); 
					elsif address = "0000000010000110" then
						readdata(15 downto 12) <= maze_desc(14)(12); 
						readdata(11 downto 8) <= maze_desc(14)(13); 
						readdata(7 downto 4) <= maze_desc(14)(14); 
						readdata(3 downto 0) <= maze_desc(14)(15); 		
						
	--				sixteenth row of maze
					elsif address = "0000000010001000" then
						readdata(15 downto 12) <= maze_desc(15)(0); 
						readdata(11 downto 8) <= maze_desc(15)(1); 
						readdata(7 downto 4) <= maze_desc(15)(2); 
						readdata(3 downto 0) <= maze_desc(15)(3); 
					elsif address = "0000000010001010" then
						readdata(15 downto 12) <= maze_desc(15)(4); 
						readdata(11 downto 8) <= maze_desc(15)(5); 
						readdata(7 downto 4) <= maze_desc(15)(6); 
						readdata(3 downto 0) <= maze_desc(15)(7); 
					elsif address = "0000000010001100" then
						readdata(15 downto 12) <= maze_desc(15)(8); 
						readdata(11 downto 8) <= maze_desc(15)(9); 
						readdata(7 downto 4) <= maze_desc(15)(10); 
						readdata(3 downto 0) <= maze_desc(15)(11); 
					elsif address = "0000000010001110" then
						readdata(15 downto 12) <= maze_desc(15)(12); 
						readdata(11 downto 8) <= maze_desc(15)(13); 
						readdata(7 downto 4) <= maze_desc(15)(14); 
						readdata(3 downto 0) <= maze_desc(15)(15); 			
						
				
					-- top row of obj
					elsif address = "0000000010010010" then
						readdata(15 downto 12) <= obj_desc(0)(0); 
						readdata(11 downto 8) <= obj_desc(0)(1); 
						readdata(7 downto 4) <= obj_desc(0)(2); 
						readdata(3 downto 0) <= obj_desc(0)(3); 
					elsif address = "0000000010010100" then
						readdata(15 downto 12) <= obj_desc(0)(4); 
						readdata(11 downto 8) <= obj_desc(0)(5); 
						readdata(7 downto 4) <= obj_desc(0)(6); 
						readdata(3 downto 0) <= obj_desc(0)(7); 
					elsif address = "0000000010010110" then
						readdata(15 downto 12) <= obj_desc(0)(8); 
						readdata(11 downto 8) <= obj_desc(0)(9); 
						readdata(7 downto 4) <= obj_desc(0)(10); 
						readdata(3 downto 0) <= obj_desc(0)(11); 
					elsif address = "0000000010011000" then
						readdata(15 downto 12) <= obj_desc(0)(12); 
						readdata(11 downto 8) <= obj_desc(0)(13); 
						readdata(7 downto 4) <= obj_desc(0)(14); 
						readdata(3 downto 0) <= obj_desc(0)(15); 
					
					-- second row of obj
					elsif address = "0000000010011010" then
						readdata(15 downto 12) <= obj_desc(1)(0); 
						readdata(11 downto 8) <= obj_desc(1)(1); 
						readdata(7 downto 4) <= obj_desc(1)(2); 
						readdata(3 downto 0) <= obj_desc(1)(3); 
					elsif address = "0000000010011100" then
						readdata(15 downto 12) <= obj_desc(1)(4); 
						readdata(11 downto 8) <= obj_desc(1)(5); 
						readdata(7 downto 4) <= obj_desc(1)(6); 
						readdata(3 downto 0) <= obj_desc(1)(7); 
					elsif address = "0000000010011110" then
						readdata(15 downto 12) <= obj_desc(1)(8); 
						readdata(11 downto 8) <= obj_desc(1)(9); 
						readdata(7 downto 4) <= obj_desc(1)(10); 
						readdata(3 downto 0) <= obj_desc(1)(11); 
					elsif address = "0000000010100000" then
						readdata(15 downto 12) <= obj_desc(1)(12); 
						readdata(11 downto 8) <= obj_desc(1)(13); 
						readdata(7 downto 4) <= obj_desc(1)(14); 
						readdata(3 downto 0) <= obj_desc(1)(15); 
						
					-- third row of obj
					elsif address = "0000000010100010" then
						readdata(15 downto 12) <= obj_desc(2)(0); 
						readdata(11 downto 8) <= obj_desc(2)(1); 
						readdata(7 downto 4) <= obj_desc(2)(2); 
						readdata(3 downto 0) <= obj_desc(2)(3); 
					elsif address = "0000000010100100" then
						readdata(15 downto 12) <= obj_desc(2)(4); 
						readdata(11 downto 8) <= obj_desc(2)(5); 
						readdata(7 downto 4) <= obj_desc(2)(6); 
						readdata(3 downto 0) <= obj_desc(2)(7); 
					elsif address = "0000000010100110" then
						readdata(15 downto 12) <= obj_desc(2)(8); 
						readdata(11 downto 8) <= obj_desc(2)(9); 
						readdata(7 downto 4) <= obj_desc(2)(10); 
						readdata(3 downto 0) <= obj_desc(2)(11); 
					elsif address = "0000000010101000" then
						readdata(15 downto 12) <= obj_desc(2)(12); 
						readdata(11 downto 8) <= obj_desc(2)(13); 
						readdata(7 downto 4) <= obj_desc(2)(14); 
						readdata(3 downto 0) <= obj_desc(2)(15); 
						
					-- fourth row of obj
					elsif address = "0000000010101010" then
						readdata(15 downto 12) <= obj_desc(3)(0); 
						readdata(11 downto 8) <= obj_desc(3)(1); 
						readdata(7 downto 4) <= obj_desc(3)(2); 
						readdata(3 downto 0) <= obj_desc(3)(3); 
					elsif address = "0000000010101100" then
						readdata(15 downto 12) <= obj_desc(3)(4); 
						readdata(11 downto 8) <= obj_desc(3)(5); 
						readdata(7 downto 4) <= obj_desc(3)(6); 
						readdata(3 downto 0) <= obj_desc(3)(7); 
					elsif address = "0000000010101110" then
						readdata(15 downto 12) <= obj_desc(3)(8); 
						readdata(11 downto 8) <= obj_desc(3)(9); 
						readdata(7 downto 4) <= obj_desc(3)(10); 
						readdata(3 downto 0) <= obj_desc(3)(11); 
					elsif address = "0000000010110000" then
						readdata(15 downto 12) <= obj_desc(3)(12); 
						readdata(11 downto 8) <= obj_desc(3)(13); 
						readdata(7 downto 4) <= obj_desc(3)(14); 
						readdata(3 downto 0) <= obj_desc(3)(15); 	
					
					-- fifth row of obj
					elsif address = "0000000010110010" then
						readdata(15 downto 12) <= obj_desc(4)(0); 
						readdata(11 downto 8) <= obj_desc(4)(1); 
						readdata(7 downto 4) <= obj_desc(4)(2); 
						readdata(3 downto 0) <= obj_desc(4)(3); 
					elsif address = "0000000010110100" then
						readdata(15 downto 12) <= obj_desc(4)(4); 
						readdata(11 downto 8) <= obj_desc(4)(5); 
						readdata(7 downto 4) <= obj_desc(4)(6); 
						readdata(3 downto 0) <= obj_desc(4)(7); 
					elsif address = "0000000010110110" then
						readdata(15 downto 12) <= obj_desc(4)(8);
						readdata(11 downto 8) <= obj_desc(4)(9); 
						readdata(7 downto 4) <= obj_desc(4)(10); 
						readdata(3 downto 0) <= obj_desc(4)(11); 
					elsif address = "0000000010111000" then
						readdata(15 downto 12) <= obj_desc(4)(12); 
						readdata(11 downto 8) <= obj_desc(4)(13); 
						readdata(7 downto 4) <= obj_desc(4)(14);
						readdata(3 downto 0) <= obj_desc(4)(15); 	
					
					-- sixth row of obj
					elsif address = "0000000010111010" then
						readdata(15 downto 12) <= obj_desc(5)(0); 
						readdata(11 downto 8) <= obj_desc(5)(1); 
						readdata(7 downto 4) <= obj_desc(5)(2); 
						readdata(3 downto 0) <= obj_desc(5)(3); 
					elsif address = "0000000010111100" then
						readdata(15 downto 12) <= obj_desc(5)(4); 
						readdata(11 downto 8) <= obj_desc(5)(5); 
						readdata(7 downto 4) <= obj_desc(5)(6); 
						readdata(3 downto 0) <= obj_desc(5)(7); 
					elsif address = "0000000010111110" then
						readdata(15 downto 12) <= obj_desc(5)(8); 
						readdata(11 downto 8) <= obj_desc(5)(9); 
						readdata(7 downto 4) <= obj_desc(5)(10); 
						readdata(3 downto 0) <= obj_desc(5)(11); 
					elsif address = "0000000011000000" then
						readdata(15 downto 12) <= obj_desc(5)(12); 
						readdata(11 downto 8) <= obj_desc(5)(13); 
						readdata(7 downto 4) <= obj_desc(5)(14); 
						readdata(3 downto 0) <= obj_desc(5)(15); 	
						
					-- seventh row of obj
					elsif address = "0000000011000010" then
						readdata(15 downto 12) <= obj_desc(6)(0); 
						readdata(11 downto 8) <= obj_desc(6)(1); 
						readdata(7 downto 4) <= obj_desc(6)(2); 
						readdata(3 downto 0) <= obj_desc(6)(3); 
					elsif address = "0000000011000100" then
						readdata(15 downto 12) <= obj_desc(6)(4); 
						readdata(11 downto 8) <= obj_desc(6)(5); 
						readdata(7 downto 4) <= obj_desc(6)(6); 
						readdata(3 downto 0) <= obj_desc(6)(7); 
					elsif address = "0000000011000110" then
						readdata(15 downto 12) <= obj_desc(6)(8); 
						readdata(11 downto 8) <= obj_desc(6)(9); 
						readdata(7 downto 4) <= obj_desc(6)(10); 
						readdata(3 downto 0) <= obj_desc(6)(11); 
					elsif address = "0000000011001000" then
						readdata(15 downto 12) <= obj_desc(6)(12); 
						readdata(11 downto 8) <= obj_desc(6)(13); 
						readdata(7 downto 4) <= obj_desc(6)(14); 
						readdata(3 downto 0) <= obj_desc(6)(15); 
					
					-- eigth row of obj
					elsif address = "0000000011001010" then
						readdata(15 downto 12) <= obj_desc(7)(0); 
						readdata(11 downto 8) <= obj_desc(7)(1); 
						readdata(7 downto 4) <= obj_desc(7)(2); 
						readdata(3 downto 0) <= obj_desc(7)(3); 
					elsif address = "0000000011001100" then
						readdata(15 downto 12) <= obj_desc(7)(4); 
						readdata(11 downto 8) <= obj_desc(7)(5); 
						readdata(7 downto 4) <= obj_desc(7)(6); 
						readdata(3 downto 0) <= obj_desc(7)(7); 
					elsif address = "0000000011001110" then
						readdata(15 downto 12) <= obj_desc(7)(8); 
						readdata(11 downto 8) <= obj_desc(7)(9); 
						readdata(7 downto 4) <= obj_desc(7)(10); 
						readdata(3 downto 0) <= obj_desc(7)(11);
					elsif address = "0000000011010000" then
						readdata(15 downto 12) <= obj_desc(7)(12); 
						readdata(11 downto 8) <= obj_desc(7)(13); 
						readdata(7 downto 4) <= obj_desc(7)(14); 
						readdata(3 downto 0) <= obj_desc(7)(15); 
					
					-- ninth row of obj
					elsif address = "0000000011010010" then
						readdata(15 downto 12) <= obj_desc(8)(0); 
						readdata(11 downto 8) <= obj_desc(8)(1); 
						readdata(7 downto 4) <= obj_desc(8)(2); 
						readdata(3 downto 0) <= obj_desc(8)(3); 
					elsif address = "0000000011010100" then
						readdata(15 downto 12) <= obj_desc(8)(4); 
						readdata(11 downto 8) <= obj_desc(8)(5); 
						readdata(7 downto 4) <= obj_desc(8)(6); 
						readdata(3 downto 0) <= obj_desc(8)(7); 
					elsif address = "0000000011010110" then
						readdata(15 downto 12) <= obj_desc(8)(8); 
						readdata(11 downto 8) <= obj_desc(8)(9); 
						readdata(7 downto 4) <= obj_desc(8)(10); 
						readdata(3 downto 0) <= obj_desc(8)(11); 
					elsif address = "0000000011011000" then
						readdata(15 downto 12) <= obj_desc(8)(12); 
						readdata(11 downto 8) <= obj_desc(8)(13); 
						readdata(7 downto 4) <= obj_desc(8)(14); 
						readdata(3 downto 0) <= obj_desc(8)(15); 
						
					-- tenth row of obj
					elsif address = "0000000011011010" then
						readdata(15 downto 12) <= obj_desc(9)(0); 
						readdata(11 downto 8) <= obj_desc(9)(1); 
						readdata(7 downto 4) <= obj_desc(9)(2); 
						readdata(3 downto 0) <= obj_desc(9)(3); 
					elsif address = "0000000011011100" then
						readdata (15 downto 12)<= obj_desc(9)(4); 
						readdata(11 downto 8) <= obj_desc(9)(5); 
						readdata(7 downto 4) <= obj_desc(9)(6); 
						readdata(3 downto 0) <= obj_desc(9)(7); 
					elsif address = "0000000011011110" then
						readdata(15 downto 12) <= obj_desc(9)(8); 
						readdata(11 downto 8) <= obj_desc(9)(9); 
						readdata(7 downto 4) <= obj_desc(9)(10); 
						readdata(3 downto 0) <= obj_desc(9)(11); 
					elsif address = "0000000011100000" then
						readdata(15 downto 12) <= obj_desc(9)(12); 
						readdata(11 downto 8) <= obj_desc(9)(13); 
						readdata(7 downto 4) <= obj_desc(9)(14); 
						readdata(3 downto 0) <= obj_desc(9)(15); 
						
					-- eleventh row of obj
					elsif address = "0000000011100010" then
						readdata(15 downto 12) <= obj_desc(10)(0); 
						readdata(11 downto 8) <= obj_desc(10)(1); 
						readdata(7 downto 4) <= obj_desc(10)(2); 
						readdata(3 downto 0) <= obj_desc(10)(3); 
					elsif address = "0000000011100100" then
						readdata(15 downto 12) <= obj_desc(10)(4); 
						readdata(11 downto 8) <= obj_desc(10)(5); 
						readdata(7 downto 4) <= obj_desc(10)(6); 
						readdata(3 downto 0) <= obj_desc(10)(7); 
					elsif address = "0000000011100110" then
						readdata(15 downto 12) <= obj_desc(10)(8); 
						readdata(11 downto 8) <= obj_desc(10)(9); 
						readdata(7 downto 4) <= obj_desc(10)(10); 
						readdata(3 downto 0) <= obj_desc(10)(11); 
					elsif address = "0000000011101000" then
						readdata(15 downto 12) <= obj_desc(10)(12); 
						readdata(11 downto 8) <= obj_desc(10)(13); 
						readdata(7 downto 4) <= obj_desc(10)(14); 
						readdata(3 downto 0) <= obj_desc(10)(15); 
						
					-- twelfth row of obj
					elsif address = "0000000011101010" then
						readdata(15 downto 12) <= obj_desc(11)(0); 
						readdata(11 downto 8) <= obj_desc(11)(1); 
						readdata(7 downto 4) <= obj_desc(11)(2); 
						readdata(3 downto 0) <= obj_desc(11)(3); 
					elsif address = "0000000011101100" then
						readdata(15 downto 12) <= obj_desc(11)(4); 
						readdata(11 downto 8) <= obj_desc(11)(5); 
						readdata(7 downto 4) <= obj_desc(11)(6); 
						readdata(3 downto 0) <= obj_desc(11)(7); 
					elsif address = "0000000011101110" then
						readdata(15 downto 12) <= obj_desc(11)(8); 
						readdata(11 downto 8) <= obj_desc(11)(9); 
						readdata(7 downto 4) <= obj_desc(11)(10); 
						readdata(3 downto 0) <= obj_desc(11)(11); 
					elsif address = "0000000011110000" then
						readdata(15 downto 12) <= obj_desc(11)(12);
						readdata(11 downto 8) <= obj_desc(11)(13); 
						readdata(7 downto 4) <= obj_desc(11)(14); 
						readdata(3 downto 0) <= obj_desc(11)(15); 		
						
	--				 thirteenth row of obj
					elsif address = "0000000011110010" then
						readdata(15 downto 12) <= obj_desc(12)(0); 
						readdata(11 downto 8) <= obj_desc(12)(1); 
						readdata(7 downto 4) <= obj_desc(12)(2); 
						readdata(3 downto 0) <= obj_desc(12)(3); 
					elsif address = "0000000011110100" then
						readdata(15 downto 12) <= obj_desc(12)(4); 
						readdata(11 downto 8) <= obj_desc(12)(5); 
						readdata(7 downto 4) <= obj_desc(12)(6); 
						readdata(3 downto 0) <= obj_desc(12)(7); 
					elsif address = "0000000011110110" then
						readdata(15 downto 12) <= obj_desc(12)(8); 
						readdata(11 downto 8) <= obj_desc(12)(9); 
						readdata(7 downto 4) <= obj_desc(12)(10); 
						readdata(3 downto 0) <= obj_desc(12)(11); 
					elsif address = "0000000011111000" then
						readdata(15 downto 12) <= obj_desc(12)(12); 
						readdata(11 downto 8) <= obj_desc(12)(13); 
						readdata(7 downto 4) <= obj_desc(12)(14); 
						readdata(3 downto 0) <= obj_desc(12)(15); 			
						
	--				 fourteenth row of obj
					elsif address = "0000000011111010" then
						readdata(15 downto 12) <= obj_desc(13)(0); 
						readdata(11 downto 8) <= obj_desc(13)(1); 
						readdata(7 downto 4) <= obj_desc(13)(2); 
						readdata(3 downto 0) <= obj_desc(13)(3); 
					elsif address = "0000000011111100" then
						readdata(15 downto 12) <= obj_desc(13)(4); 
						readdata(11 downto 8) <= obj_desc(13)(5); 
						readdata(7 downto 4) <= obj_desc(13)(6); 
						readdata(3 downto 0) <= obj_desc(13)(7);
					elsif address = "0000000011111110" then
						readdata(15 downto 12) <= obj_desc(13)(8); 
						readdata(11 downto 8) <= obj_desc(13)(9); 
						readdata(7 downto 4) <= obj_desc(13)(10); 
						readdata(3 downto 0) <= obj_desc(13)(11); 
					elsif address = "0000000100000000" then
						readdata(15 downto 12) <= obj_desc(13)(12); 
						readdata(11 downto 8) <= obj_desc(13)(13); 
						readdata(7 downto 4) <= obj_desc(13)(14); 
						readdata(3 downto 0) <= obj_desc(13)(15); 		
						
	--				 fifteenth row of obj
					elsif address = "0000000100000010" then
						readdata(15 downto 12) <= obj_desc(14)(0); 
						readdata(11 downto 8) <= obj_desc(14)(1); 
						readdata(7 downto 4) <= obj_desc(14)(2); 
						readdata(3 downto 0) <= obj_desc(14)(3); 
					elsif address = "0000000100000100" then
						readdata(15 downto 12) <= obj_desc(14)(4); 
						readdata(11 downto 8) <= obj_desc(14)(5); 
						readdata(7 downto 4) <= obj_desc(14)(6); 
						readdata(3 downto 0) <= obj_desc(14)(7); 
					elsif address = "0000000100000110" then
						readdata(15 downto 12) <= obj_desc(14)(8); 
						readdata(11 downto 8) <= obj_desc(14)(9); 
						readdata(7 downto 4) <= obj_desc(14)(10); 
						readdata(3 downto 0) <= obj_desc(14)(11); 
					elsif address = "0000000100001000" then
						readdata(15 downto 12) <= obj_desc(14)(12); 
						readdata(11 downto 8) <= obj_desc(14)(13); 
						readdata(7 downto 4) <= obj_desc(14)(14); 
						readdata(3 downto 0) <= obj_desc(14)(15); 		
						
	--				sixteenth row of obj
					elsif address = "0000000100001010" then
						readdata(15 downto 12) <= obj_desc(15)(0); 
						readdata(11 downto 8) <= obj_desc(15)(1); 
						readdata(7 downto 4) <= obj_desc(15)(2); 
						readdata(3 downto 0) <= obj_desc(15)(3); 
					elsif address = "0000000100001100" then
						readdata(15 downto 12) <= obj_desc(15)(4); 
						readdata(11 downto 8) <= obj_desc(15)(5); 
						readdata(7 downto 4) <= obj_desc(15)(6); 
						readdata(3 downto 0) <= obj_desc(15)(7); 
					elsif address = "0000000100001110" then
						readdata(15 downto 12) <= obj_desc(15)(8); 
						readdata(11 downto 8) <= obj_desc(15)(9); 
						readdata(7 downto 4) <= obj_desc(15)(10); 
						readdata(3 downto 0) <= obj_desc(15)(11); 
					elsif address = "0000000100010000" then
						readdata(15 downto 12) <= obj_desc(15)(12); 
						readdata(11 downto 8) <= obj_desc(15)(13); 
						readdata(7 downto 4) <= obj_desc(15)(14); 
						readdata(3 downto 0) <= obj_desc(15)(15); 		
						
					else
						 readdata <= foo;
						
					end if;		
				end if;
				
			end if;
	end if;
 end process GetData;

DoIncPos : process (clk)
begin
	if rising_edge(clk) then
		if EndOfField = '1' then
			maze_x_offset <= tmp_maze_x_offset;
			maze_y_offset <= tmp_maze_y_offset;
			cur_xpos_pix <= tmp_cur_xpos_pix;
			cur_ypos_pix <= tmp_cur_ypos_pix;
			which_char_tile <= tmp_char_tile;
		end if;
	end if;
end process DoIncPos;
		
end rtl;
